# coding: utf-8

from __future__ import unicode_literals

from contextlib import contextmanager
from threading import Lock
import random
import string  # pylint:disable=deprecated-module
import sys

import six
from six.moves.urllib.parse import urlencode, urlunsplit  # pylint:disable=import-error,no-name-in-module

from boxsdk.config import API
from boxsdk.exception import BoxOAuthException
from boxsdk.network.default_network import DefaultNetwork
from boxsdk.object.base_api_json_object import BaseAPIJSONObject
from boxsdk.util.text_enum import TextEnum


class TokenScope(TextEnum):
    """ Scopes used for a downscope token request.

    See https://developer.box.com/reference#token-exchange.
    """
    ITEM_READ = 'item_read'
    ITEM_READWRITE = 'item_readwrite'
    ITEM_PREVIEW = 'item_preview'
    ITEM_UPLOAD = 'item_upload'
    ITEM_SHARE = 'item_share'
    ITEM_DELETE = 'item_delete'
    ITEM_DOWNLOAD = 'item_download'


class TokenResponse(BaseAPIJSONObject):
    """ Represents the response for a token request. """
    pass


class OAuth2(object):
    """
    Responsible for handling OAuth2 for the Box API. Can authenticate and refresh tokens.

    Can be used as a closeable resource, similar to a file. When `close()` is
    called, the current tokens are revoked, and the object is put into a state
    where it can no longer request new tokens. This action can also be managed
    with the `closing()` context manager method.
    """

    def __init__(
            self,
            client_id,
            client_secret,
            store_tokens=None,
            box_device_id='0',
            box_device_name='',
            access_token=None,
            refresh_token=None,
            network_layer=None,
            refresh_lock=None,
    ):
        """
        :param client_id:
            Box API key used for identifying the application the user is authenticating with.
        :type client_id:
            `unicode`
        :param client_secret:
            Box API secret used for making OAuth2 requests.
        :type client_secret:
            `unicode`
        :param store_tokens:
            Optional callback for getting access to tokens for storing them.
        :type store_tokens:
            `callable`
        :param box_device_id:
            Optional unique ID of this device. Used for applications that want to support device-pinning.
        :type box_device_id:
            `unicode`
        :param box_device_name:
            Optional human readable name for this device.
        :type box_device_name:
            `unicode`
        :param access_token:
            Access token to use for auth until it expires.
        :type access_token:
            `unicode`
        :param refresh_token:
            Refresh token to use for auth until it expires or is used.
        :type refresh_token:
            `unicode`
        :param network_layer:
            If specified, use it to make network requests. If not, the default network implementation will be used.
        :type network_layer:
            :class:`Network`
        :param refresh_lock:
            Lock used to synchronize token refresh. If not specified, then a :class:`threading.Lock` will be used.
        :type refresh_lock:
            Context Manager
        """
        self._client_id = client_id
        self._client_secret = client_secret
        self._store_tokens_callback = store_tokens
        self._access_token = access_token
        self._refresh_token = refresh_token
        self._network_layer = network_layer if network_layer else DefaultNetwork()
        self._refresh_lock = refresh_lock or Lock()
        self._box_device_id = box_device_id
        self._box_device_name = box_device_name
        self._closed = False

    @property
    def access_token(self):
        """
        Get the current access token.

        :return:
            current access token
        :rtype:
            `unicode`
        """
        return self._access_token

    @property
    def closed(self):
        """True iff the auth object has been closed.

        When in the closed state, it can no longer request new tokens.

        :rtype:   `bool`
        """
        return self._closed

    def get_authorization_url(self, redirect_url):
        """
        Get the authorization url based on the client id and the redirect url passed in

        :param redirect_url:
            An HTTPS URI or custom URL scheme where the response will be redirected. Optional if the redirect URI is
            registered with Box already.
        :type redirect_url:
            `unicode` or None
        :return:
            A tuple of the URL of Box's authorization page and the CSRF token.
            This is the URL that your application should forward the user to in first leg of OAuth 2.
        :rtype:
            (`unicode`, `unicode`)
        """
        csrf_token = self._get_state_csrf_token()
        # For the query string parameters, use a sequence of two-element
        # tuples, rather than a dictionary, in order to get a consistent and
        # predictable order of parameters in the output of `urlencode()`.
        params = [
            ('state', csrf_token),
            ('response_type', 'code'),
            ('client_id', self._client_id),
        ]
        if redirect_url:
            params.append(('redirect_uri', redirect_url))
        # `urlencode()` doesn't work with non-ASCII unicode characters, so
        # encode the parameters as ASCII bytes.
        params = [(key.encode('utf-8'), value.encode('utf-8')) for (key, value) in params]
        query_string = urlencode(params)
        return urlunsplit(('', '', API.OAUTH2_AUTHORIZE_URL, query_string, '')), csrf_token

    def authenticate(self, auth_code):
        """
        Send token request and return the access_token, refresh_token tuple. The access token and refresh token will be
        stored by calling the `store_tokens` callback if provided in __init__.

        :param auth_code:
            An authorization code you retrieved in the first leg of OAuth 2.
        :type auth_code:
            `unicode` or None

        :return:
            (access_token, refresh_token)
        :rtype:
            (`unicode`, `unicode`)
        """
        data = {
            'grant_type': 'authorization_code',
            'code': auth_code,
            'client_id': self._client_id,
            'client_secret': self._client_secret,
        }
        if self._box_device_id:
            data['box_device_id'] = self._box_device_id
        if self._box_device_name:
            data['box_device_name'] = self._box_device_name
        return self.send_token_request(data, access_token=None)

    def _refresh(self, access_token):
        data = {
            'grant_type': 'refresh_token',
            'refresh_token': self._refresh_token,
            'client_id': self._client_id,
            'client_secret': self._client_secret,
        }
        if self._box_device_id:
            data['box_device_id'] = self._box_device_id
        if self._box_device_name:
            data['box_device_name'] = self._box_device_name

        return self.send_token_request(data, access_token)

    def _get_tokens(self):
        """
        Get the current access and refresh tokens.

        This is a protected method that can be overridden to look up tokens
        from an external source (the inverse of the `store_tokens` callback).

        This method does not need to update this object's private token
        attributes. Its caller in :class:`OAuth2` is responsible for that.

        :return:
            Tuple containing the current access token and refresh token.
            One or both of them may be `None`, if they aren't set.
        :rtype:
            `tuple` of ((`unicode` or `None`), (`unicode` or `None`))
        """
        return self._access_token, self._refresh_token

    def refresh(self, access_token_to_refresh):
        """
        Refresh the access token and the refresh token and return the access_token, refresh_token tuple. The access
        token and refresh token will be stored by calling the `store_tokens` callback if provided in __init__.

        :param access_token_to_refresh:
            The expired access token, which needs to be refreshed.
            Pass `None` if you don't have the access token.
        :type access_token_to_refresh:
            `unicode` or `None`
        :return:
            Tuple containing the new access token and refresh token.
            The refresh token may be `None`, if the authentication scheme
            doesn't use one, or keeps it hidden from this client.
        :rtype:
            `tuple` of (`unicode`, (`unicode` or `None`))
        """
        self._check_closed()
        with self._refresh_lock:
            self._check_closed()
            access_token, refresh_token = self._get_and_update_current_tokens()
            # The lock here is for handling that case that multiple requests fail, due to access token expired, at the
            # same time to avoid multiple session renewals.
            if (access_token is None) or (access_token_to_refresh == access_token):
                # If the active access token is the same as the token that needs to
                # be refreshed, or if we don't currently have any active access
                # token, we make the request to refresh the token.
                access_token, refresh_token = self._refresh(access_token_to_refresh)
            # Else, if the active access token (self._access_token) is not the same as the token needs to be refreshed,
            # it means the expired token has already been refreshed. Simply return the current active tokens.
            return access_token, refresh_token

    @staticmethod
    def _get_state_csrf_token():
        """ Generate a random state CSRF token to be used in the authorization url.
        Example: box_csrf_token_Iijw9aU31sNdgiQu

        :return:
            The security token
        :rtype:
            `unicode`
        """
        system_random = random.SystemRandom()
        ascii_alphabet = string.ascii_letters + string.digits
        ascii_len = len(ascii_alphabet)
        return 'box_csrf_token_' + ''.join(ascii_alphabet[int(system_random.random() * ascii_len)] for _ in range(16))

    def _store_tokens(self, access_token, refresh_token):
        self._update_current_tokens(access_token, refresh_token)
        if self._store_tokens_callback is not None:
            self._store_tokens_callback(access_token, refresh_token)

    def _get_and_update_current_tokens(self):
        """Get the current access and refresh tokens, while also storing them in this object's private attributes.

        :return:
            Same as for :meth:`_get_tokens()`.
        """
        tokens = self._get_tokens()
        self._update_current_tokens(*tokens)
        return tokens

    def _update_current_tokens(self, access_token, refresh_token):
        """Store the latest tokens in this object's private attributes.

        :param access_token:
            The latest access token.
            May be `None`, if it hasn't been provided.
        :type access_token:
            `unicode` or `None`
        :param refresh_token:
            The latest refresh token.
            May be `None`, if the authentication scheme doesn't use one, or if
            it hasn't been provided.
        :type refresh_token:
            `unicode` or `None`
        """
        self._access_token, self._refresh_token = access_token, refresh_token

    def _execute_token_request(self, data, access_token, expect_refresh_token=True):
        """
        Send the request to acquire or refresh an access token.

        :param data:
            Dictionary containing the request parameters as specified by the Box API.
        :type data:
            `dict`
        :param access_token:
            The current access token.
        :type access_token:
            `unicode` or None
        :return:
            The response for the token request.
        :rtype:
            :class:`TokenResponse`
        """
        self._check_closed()
        url = '{base_auth_url}/token'.format(base_auth_url=API.OAUTH2_API_URL)
        headers = {'content-type': 'application/x-www-form-urlencoded'}
        network_response = self._network_layer.request(
            'POST',
            url,
            data=data,
            headers=headers,
            access_token=access_token,
        )
        if not network_response.ok:
            raise BoxOAuthException(network_response.status_code, network_response.content, url, 'POST', network_response)
        try:
            token_response = TokenResponse(network_response.json())
        except ValueError:
            raise BoxOAuthException(network_response.status_code, network_response.content, url, 'POST', network_response)

        if ('access_token' not in token_response) or (expect_refresh_token and 'refresh_token' not in token_response):
            raise BoxOAuthException(network_response.status_code, network_response.content, url, 'POST', network_response)

        return token_response

    def send_token_request(self, data, access_token, expect_refresh_token=True):
        """
        Send the request to acquire or refresh an access token, and store the tokens.

        :param data:
            Dictionary containing the request parameters as specified by the Box API.
        :type data:
            `dict`
        :param access_token:
            The current access token.
        :type access_token:
            `unicode` or None
        :return:
            The access token and refresh token.
        :rtype:
            (`unicode`, `unicode`)
        """
        token_response = self._execute_token_request(data, access_token, expect_refresh_token)
        # pylint:disable=no-member
        refresh_token = token_response.refresh_token if 'refresh_token' in token_response else None
        self._store_tokens(token_response.access_token, refresh_token)
        return self._access_token, self._refresh_token

    def revoke(self):
        """
        Revoke the authorization for the current access/refresh token pair.
        """
        with self._refresh_lock:
            access_token, refresh_token = self._get_and_update_current_tokens()
            token_to_revoke = access_token or refresh_token
            if token_to_revoke is None:
                return
            url = '{base_auth_url}/revoke'.format(base_auth_url=API.OAUTH2_API_URL)
            network_response = self._network_layer.request(
                'POST',
                url,
                data={
                    'client_id': self._client_id,
                    'client_secret': self._client_secret,
                    'token': token_to_revoke,
                },
                access_token=access_token,
            )
            if not network_response.ok:
                raise BoxOAuthException(network_response.status_code, network_response.content, url, 'POST', network_response)
            self._store_tokens(None, None)

    def close(self, revoke=True):
        """Close the auth object.

        After this action is performed, the auth object can no longer request
        new tokens.

        This method may be called even if the auth object is already closed.

        :param revoke:
            (optional) Whether the current tokens should be revoked, via `revoke()`.
            Defaults to `True` as a security precaution, so that the tokens aren't usable
            by any adversaries after you are done with them.
            Note that the revoke isn't guaranteed to succeed (the network connection might
            fail, or the API call might respond with a non-200 HTTP response), so this
            isn't a fool-proof security mechanism.
            If the revoke fails, an exception is raised.
            The auth object is still considered to be closed, even if the revoke fails.
        :type revoke:   `bool`
        """
        self._closed = True
        if revoke:
            self.revoke()

    @contextmanager
    def closing(self, **close_kwargs):
        """Context manager to close the auth object on exit.

        The behavior is somewhat similar to `contextlib.closing(self)`, but has
        some differences.

        The context manager cannot be entered if the auth object is closed.

        If a non-`Exception` (e.g. `KeyboardInterrupt`) is caught from the
        block, this context manager prioritizes re-raising the exception as
        fast as possible, without blocking. Thus, in this case, the tokens will
        not be revoked, even if `revoke=True` was passed to this method.

        If exceptions are raised both from the block and from `close()`, the
        exception from the block will be reraised, and the exception from
        `close()` will be swallowed. The assumption is that the exception from
        the block is more relevant to the client, especially since the revoke
        can fail if the network is unavailable.

        :param **close_kwargs:  Keyword arguments to pass to `close()`.
        """
        self._check_closed()
        exc_infos = []

        # pylint:disable=broad-except
        try:
            yield self
        except Exception:
            exc_infos.append(sys.exc_info())
        except BaseException:
            exc_infos.append(sys.exc_info())
            close_kwargs['revoke'] = False

        try:
            self.close(**close_kwargs)
        except Exception:
            exc_infos.append(sys.exc_info())

        if exc_infos:
            six.reraise(*exc_infos[0])

    def _check_closed(self):
        if self.closed:
            raise ValueError("operation on a closed auth object")



# coding: utf-8

from __future__ import unicode_literals, absolute_import

from .cloneable import Cloneable


class BaseEndpoint(Cloneable):
    """A Box API endpoint."""

    def __init__(self, session, **kwargs):
        """
        :param session:
            The Box session used to make requests.
        :type session:
            :class:`BoxSession`
        :param kwargs:
            Keyword arguments for base class constructors.
        :type kwargs:
            `dict`
        """
        super(BaseEndpoint, self).__init__(**kwargs)
        self._session = session

    @property
    def session(self):
        """
        Get the :class:`BoxSession` instance the object is using.

        :rtype:
            :class:`BoxSession`
        """
        return self._session

    @property
    def translator(self):
        """The translator used for translating Box API JSON responses into `BaseAPIJSONObject` smart objects.

        :rtype:   :class:`Translator`
        """
        return self._session.translator

    def get_url(self, endpoint, *args):
        """
        Return the URL used to access the endpoint.

        :param endpoint:
            The name of the endpoint.
        :type endpoint:
            `url`
        :param args:
            Additional parts of the endpoint URL.
        :type args:
            `Iterable`
        :rtype:
            `unicode`
        """
        # pylint:disable=no-self-use
        return self._session.get_url(endpoint, *args)

    def clone(self, session=None):
        """
        Returns a copy of this cloneable object using the specified session.

        :param session:
            The Box session used to make requests.
        :type session:
            :class:`BoxSession`
        """
        return self.__class__(session or self._session)



# coding: utf-8

from __future__ import unicode_literals, absolute_import
import pytest

from boxsdk.object.api_json_object import APIJSONObject


@pytest.fixture(params=[{'foo': 'bar'}, {'a': {'b': 'c'}}])
def api_json_object(request):
    return request.param, APIJSONObject(request.param)


def test_len(api_json_object):
    dictionary, test_object = api_json_object
    assert len(dictionary) == len(test_object)


def test_api_json_object_dict(api_json_object):
    dictionary, test_object = api_json_object
    assert dictionary == test_object



# coding: utf-8

from __future__ import absolute_import, unicode_literals

from mock import DEFAULT, Mock, patch
import pytest
from requests import Response

from boxsdk.network.default_network import DefaultNetworkResponse, DefaultNetwork


@pytest.fixture
def make_network_request(make_network_request, request_response, http_verb):
    # pylint:disable=unused-argument
    # Need to list `http_verb`, even though it isn't used in this override,
    # because of <https://github.com/pytest-dev/pytest/issues/1953>.

    def _make_network_request(*args, **kwargs):
        response = make_network_request(*args, **kwargs)
        assert response.request_response is request_response
        return response

    return _make_network_request


def test_default_network_response_properties_pass_through_to_session_response_properties(access_token):
    mock_session_response = Mock(Response)
    mock_session_response.status_code = 200
    mock_session_response.headers = {}
    mock_session_response.raw = Mock()
    network_reponse = DefaultNetworkResponse(mock_session_response, access_token)
    assert network_reponse.json() == mock_session_response.json()
    assert network_reponse.content == mock_session_response.content
    assert network_reponse.ok == mock_session_response.ok
    assert network_reponse.status_code == mock_session_response.status_code
    assert network_reponse.headers == mock_session_response.headers
    assert network_reponse.response_as_stream == mock_session_response.raw
    assert network_reponse.access_token_used == access_token


def test_default_network_request(make_network_request):
    # pylint:disable=redefined-outer-name
    default_network = DefaultNetwork()
    make_network_request(default_network, custom_kwargs='test')


@pytest.mark.parametrize('delay', (0, 1))
def test_default_network_retry_after_sleeps(delay):
    default_network = DefaultNetwork()
    retry_call = Mock()
    mock_sleep = Mock()
    with patch('boxsdk.network.default_network.time.sleep', mock_sleep):
        default_network.retry_after(delay, retry_call, DEFAULT, kwarg=DEFAULT)
    mock_sleep.assert_called_once_with(delay)
    retry_call.assert_called_once_with(DEFAULT, kwarg=DEFAULT)


def test_network_response_constructor(make_network_request):
    assert DefaultNetwork().network_response_constructor is DefaultNetworkResponse

    class DefaultNetworkResponseSubclass(DefaultNetworkResponse):
        pass

    class DefaultNetworkSubclass(DefaultNetwork):
        @property
        def network_response_constructor(self):
            return DefaultNetworkResponseSubclass

    network = DefaultNetworkSubclass()
    response = make_network_request(network)
    assert isinstance(response, DefaultNetworkResponseSubclass)



# coding: utf-8

from __future__ import unicode_literals

from .box_object_collection import BoxObjectCollection


class LimitOffsetBasedObjectCollection(BoxObjectCollection):
    """
    An iterator of Box objects (BaseObjects) that were retrieved from a Box API endpoint that supports
    limit-offset type of pagination.

    See https://developer.box.com/reference#pagination for more details.
    """

    def __init__(
            self,
            session,
            url,
            limit=None,
            fields=None,
            additional_params=None,
            return_full_pages=False,
            offset=0,
    ):
        """
        :param offset:
            The offset index to start paging from.
        :type offset:
            `int`
        """
        super(LimitOffsetBasedObjectCollection, self).__init__(
            session,
            url,
            limit=limit,
            fields=fields,
            additional_params=additional_params,
            return_full_pages=return_full_pages,
        )
        self._offset = offset

    def _update_pointer_to_next_page(self, response_object):
        """Baseclass override."""
        total_count = response_object['total_count']

        # The API might use a lower limit than the client asked for, if the
        # client asked for a limit above the maximum limit for that endpoint.
        # The API is supposed to respond with the limit that it actually used.
        # If that is given, then use that limit for the offset calculation, and
        # also for the remainder of the paging.
        #
        # Similarly, the API reports the offset that it used. In theory, this
        # should always be the same as what was requested. But just in case, do
        # the same thing with offset.
        if 'limit' in response_object:
            self._limit, old_limit = int(response_object['limit']), self._limit

            # If the API erroneously sends a bad value for limit, we want to
            # avoid getting into an infinite chain of API calls. So abort with
            # a runtime error.
            if self._limit <= 0 < old_limit:
                self._offset = total_count  # Disable additional paging.
                raise RuntimeError('API returned limit={0}, cannot continue paging'.format(self._limit))

        if 'offset' in response_object:
            self._offset = int(response_object['offset'])

        if total_count >= self._offset + self._limit:
            self._offset += self._limit
        else:
            self._offset = total_count

    def _has_more_pages(self, response_object):
        """Baseclass override."""
        return self._offset < response_object['total_count']

    def _next_page_pointer_params(self):
        """Baseclass override."""
        return {'offset': self._offset}

    def next_pointer(self):
        """Baseclass override."""
        return self._offset



# coding: utf-8

from __future__ import unicode_literals
import sqlalchemy
import uuid
from test.functional.mock_box.db_model import DbModel


class CollaborationModel(DbModel):
    """DB Model for Box collaborations."""
    __tablename__ = 'box_collaboration'

    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True, autoincrement=True)  # pylint:disable=invalid-name
    collab_id = sqlalchemy.Column(sqlalchemy.String(32), default=lambda: uuid.uuid4().hex)
    item_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey('box_folder.id'))
    created_at = sqlalchemy.Column(sqlalchemy.DateTime)
    modified_at = sqlalchemy.Column(sqlalchemy.DateTime)
    created_by_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey('box_user.id'))
    status = sqlalchemy.Column(sqlalchemy.String(10))



# coding: utf-8

from __future__ import unicode_literals, absolute_import

from .jwt_auth import JWTAuth
from .redis_managed_oauth2 import RedisManagedOAuth2Mixin


class RedisManagedJWTAuth(RedisManagedOAuth2Mixin, JWTAuth):
    """
    JWT Auth subclass which uses Redis to manage access tokens.
    """
    def _auth_with_jwt(self, sub, sub_type):
        """
        Base class override. Returns the access token in a tuple to match the OAuth2 interface.
        """
        return super(RedisManagedJWTAuth, self)._auth_with_jwt(sub, sub_type), None



# coding: utf-8

from __future__ import unicode_literals, absolute_import
from abc import ABCMeta, abstractmethod
from six import add_metaclass
from six.moves import range   # pylint:disable=redefined-builtin
import pytest

from boxsdk.util.translator import Translator


@add_metaclass(ABCMeta)
class BoxObjectCollectionTestBase(object):
    NUM_ENTRIES = 25

    @staticmethod
    @pytest.fixture()
    def translator():
        return Translator()

    @pytest.fixture()
    def entries(self):
        all_entries = []
        for i in range(self.NUM_ENTRIES):
            all_entries.append({
                "type": "file",
                "id": str(1000 + i),
                "sequence_id": str(i),
                "etag": str(10 + i),
                "name": "file_{0}.txt".format(i),
            })
        return all_entries

    @abstractmethod
    @pytest.fixture()
    def mock_items_response(self, entries):
        raise NotImplementedError

    @abstractmethod
    @pytest.fixture()
    def mock_session(self, translator, mock_items_response):
        raise NotImplementedError

    @abstractmethod
    def _object_collection_instance(self, session, limit, return_full_pages=False, starting_pointer=None):
        """
        :type session: :class:`BoxSession`
        :type limit: `int`
        :type return_full_pages: `bool`
        :type starting_pointer: varies
        :rtype: :class:`BoxObjectCollection`
        """
        raise NotImplementedError

    @staticmethod
    def _assert_items_dict_and_objects_same(expected_items_dict, returned_item_objects):
        """
        A fixture very specific to this test class. Asserts that the list of items in dictionary form are the
        same (at least in name, and in quantity) as a list of BaseObjects.

        :param expected_items_dict: List of expected items, represented as a dictionary.
        :type expected_items_dict: `list` of `dict`
        :param returned_item_objects: List of item instances (BaseObject) returned by SUT.
        :type returned_item_objects: `list` of class:`BaseObject`
        """
        expected_num = len(expected_items_dict)
        actual_num = len(returned_item_objects)
        assert actual_num == expected_num, 'Expected {0} items, got {1}'.format(expected_num, actual_num)
        returned_item_names = [item.name for item in returned_item_objects]
        for expected_item_dict in expected_items_dict:
            assert expected_item_dict['name'] in returned_item_names, 'Missing item: {0}'.format(expected_item_dict['name'])

    @pytest.mark.parametrize('return_full_pages', (True, False))
    @pytest.mark.parametrize('limit', (1, 3, 5, NUM_ENTRIES, 1000))
    def test_object_collection_pages_through_all_entries(self, mock_session, entries, limit, return_full_pages):
        """
        Tests the basic iteration functionality of the box object collection.
        """
        object_collection = self._object_collection_instance(mock_session, limit, return_full_pages)
        iterated_items = []
        for item_or_page in object_collection:
            if return_full_pages:
                iterated_items.extend(item_or_page)
            else:
                iterated_items.append(item_or_page)
        self._assert_items_dict_and_objects_same(entries, iterated_items)

    def test_new_object_collection_starts_off_from_last_pointer(self, mock_session, entries):
        """
        Start paging with one object collection instance, and then finish paging with a new object collection
        instance, starting off from the next_pointer() of the previous instance.

        The iterated items should be identical as if it were iterated through with just one object collection
        instance.
        """
        iterated_items = []
        object_collection = self._object_collection_instance(mock_session, limit=5, return_full_pages=True)
        iterated_items.extend(object_collection.next())
        next_pointer = object_collection.next_pointer()

        # Create a new object collection, starting off where the previous collection left off.
        new_object_collection = self._object_collection_instance(
            mock_session,
            limit=5,
            return_full_pages=True,
            starting_pointer=next_pointer
        )
        for page in new_object_collection:
            iterated_items.extend(page)

        self._assert_items_dict_and_objects_same(entries, iterated_items)



# coding: utf-8

from __future__ import unicode_literals

import json

import pytest

from boxsdk.object.collaboration import CollaborationRole, CollaborationStatus


@pytest.mark.parametrize('data', [
    {},
    {'role': CollaborationRole.EDITOR},
    {'role': CollaborationRole.VIEWER},
    {'status': CollaborationStatus.ACCEPTED},
    {'status': CollaborationStatus.REJECTED},
    {'role': CollaborationRole.EDITOR, 'status': CollaborationStatus.ACCEPTED},
])
def test_update_info_returns_the_correct_response(
        test_collaboration,
        mock_box_session,
        mock_collab_response,
        data):
    # pylint:disable=protected-access
    expected_url = test_collaboration.get_url()
    mock_box_session.put.return_value = mock_collab_response
    update_response = test_collaboration.update_info(**data)
    mock_box_session.put.assert_called_once_with(
        expected_url,
        data=json.dumps(data),
        headers=None,
        params=None,
    )
    assert isinstance(update_response, test_collaboration.__class__)
    assert update_response.object_id == test_collaboration.object_id



# coding: utf-8

from __future__ import absolute_import, unicode_literals

import io
import logging

from mock import mock_open, patch, Mock
import pytest
from six import string_types

import boxsdk.util.log


_MOCK_FILEPATH = '/home/user/boxsdk.log'
_MOCK_LOG_NAME = 'boxsdk'


@pytest.fixture(params=[io.StringIO(), _MOCK_FILEPATH])
def stream_or_file(request):
    return request.param


@pytest.fixture(params=[False, True])
def debug(request):
    return request.param


@pytest.fixture
def expected_log_level(debug):
    return logging.DEBUG if debug else logging.INFO


@pytest.fixture(params=[_MOCK_LOG_NAME, None])
def name(request):
    return request.param


@pytest.fixture
def mock_logger():
    return Mock(logging.Logger)


def test_setup_logging(stream_or_file, debug, expected_log_level, name, mock_logger):
    mock_file_open = mock_open()

    with patch('logging.getLogger') as get_logger:
        with patch('logging.open', mock_file_open, create=True):
            get_logger.return_value = mock_logger
            assert boxsdk.util.log.setup_logging(stream_or_file, debug=debug, name=name) == mock_logger
            get_logger.assert_called_once_with(name)

    assert mock_logger.addHandler.call_count == 1
    assert isinstance(mock_logger.addHandler.call_args[0][0], logging.Handler)
    mock_logger.setLevel.assert_called_once_with(expected_log_level)

    if isinstance(stream_or_file, string_types):
        assert mock_file_open.call_count == 1
        assert mock_file_open.call_args[0][:2] == (stream_or_file, 'w')   # Python 3 passes additional args.



# coding: utf-8

from __future__ import unicode_literals, absolute_import

from .api_json_object import APIJSONObject


class Event(APIJSONObject):
    """Represents a single Box event."""

    _item_type = 'event'



# coding: utf-8

from __future__ import unicode_literals, absolute_import

from . import version


class API(object):
    """Configuration object containing the URLs for the Box API."""
    BASE_API_URL = 'https://api.box.com/2.0'
    UPLOAD_URL = 'https://upload.box.com/api/2.0'
    OAUTH2_API_URL = 'https://api.box.com/oauth2'  # <https://developers.box.com/docs/#oauth-2>
    OAUTH2_AUTHORIZE_URL = 'https://app.box.com/api/oauth2/authorize'  # <https://developers.box.com/docs/#oauth-2-authorize>


class Client(object):
    """Configuration object containing the user agent string."""
    VERSION = version.__version__
    USER_AGENT_STRING = 'box-python-sdk-{0}'.format(VERSION)



# coding: utf-8

from __future__ import unicode_literals, absolute_import

from six.moves import input  # pylint:disable=redefined-builtin

from .oauth2 import OAuth2


class DeveloperTokenAuth(OAuth2):
    ENTER_TOKEN_PROMPT = 'Enter developer token: '

    def __init__(self, get_new_token_callback=None, **kwargs):
        self._get_new_token = get_new_token_callback
        super(DeveloperTokenAuth, self).__init__(
            client_id=None,
            client_secret=None,
            access_token=self._refresh_developer_token(),
            **kwargs
        )

    def _refresh_developer_token(self):
        if self._get_new_token is not None:
            return self._get_new_token()

        return input(self.ENTER_TOKEN_PROMPT)

    def _refresh(self, access_token):
        """
        Base class override.
        Ask for a new developer token.
        """
        self._access_token = self._refresh_developer_token()
        return self._access_token, None

    def revoke(self):
        """
        Base class override.
        Do nothing; developer tokens can't be revoked without client ID and secret.
        """



# coding: utf-8

from __future__ import unicode_literals

from boxsdk.config import API
from .item import Item
from ..util.api_call_decorator import api_call


class File(Item):
    """Box API endpoint for interacting with files."""

    _item_type = 'file'

    def preflight_check(self, size, name=None):
        """
        Make an API call to check if the file can be updated with the new name and size of the file.

        :param size:
            The size of the file in bytes. Specify 0 for unknown file-sizes.
        :type size:
            `int`
        :param name:
            The name of the file to be updated. It's optional, if the name is not being changed.
        :type name:
            `unicode`
        :raises:
            :class:`BoxAPIException` when preflight check fails.
        """
        self._preflight_check(
            size=size,
            name=name,
            file_id=self._object_id,
        )

    def _get_accelerator_upload_url_for_update(self):
        """
        Get Accelerator upload url for updating the file.

        :return:
            The Accelerator upload url for updating the file or None if cannot get one
        :rtype:
            `unicode` or None
        """
        return self._get_accelerator_upload_url(file_id=self._object_id)

    @api_call
    def content(self):
        """
        Get the content of a file on Box.

        :returns:
            File content as bytes.
        :rtype:
            `bytes`
        """
        url = self.get_url('content')
        box_response = self._session.get(url, expect_json_response=False)
        return box_response.content

    @api_call
    def download_to(self, writeable_stream):
        """
        Download the file; write it to the given stream.

        :param writeable_stream:
            A file-like object where bytes can be written into.
        :type writeable_stream:
            `file`
        """
        url = self.get_url('content')
        box_response = self._session.get(url, expect_json_response=False, stream=True)
        for chunk in box_response.network_response.response_as_stream.stream(decode_content=True):
            writeable_stream.write(chunk)

    @api_call
    def update_contents_with_stream(
            self,
            file_stream,
            etag=None,
            preflight_check=False,
            preflight_expected_size=0,
            upload_using_accelerator=False,
    ):
        """
        Upload a new version of a file, taking the contents from the given file stream.

        :param file_stream:
            The file-like object containing the bytes
        :type file_stream:
            `file`
        :param etag:
            If specified, instruct the Box API to update the item only if the current version's etag matches.
        :type etag:
            `unicode` or None
        :param preflight_check:
            If specified, preflight check will be performed before actually uploading the file.
        :type preflight_check:
            `bool`
        :param preflight_expected_size:
            The size of the file to be uploaded in bytes, which is used for preflight check. The default value is '0',
            which means the file size is unknown.
        :type preflight_expected_size:
            `int`
        :param upload_using_accelerator:
            If specified, the upload will try to use Box Accelerator to speed up the uploads for big files.
            It will make an extra API call before the actual upload to get the Accelerator upload url, and then make
            a POST request to that url instead of the default Box upload url. It falls back to normal upload endpoint,
            if cannot get the Accelerator upload url.

            Please notice that this is a premium feature, which might not be available to your app.
        :type upload_using_accelerator:
            `bool`
        :returns:
            A new file object
        :rtype:
            :class:`File`
        :raises:
            :class:`BoxAPIException` if the specified etag doesn't match the latest version of the file or preflight
            check fails.
        """
        if preflight_check:
            self.preflight_check(size=preflight_expected_size)

        url = self.get_url('content').replace(API.BASE_API_URL, API.UPLOAD_URL)
        if upload_using_accelerator:
            accelerator_upload_url = self._get_accelerator_upload_url_for_update()
            if accelerator_upload_url:
                url = accelerator_upload_url

        files = {'file': ('unused', file_stream)}
        headers = {'If-Match': etag} if etag is not None else None
        file_response = self._session.post(url, expect_json_response=False, files=files, headers=headers).json()
        if 'entries' in file_response:
            file_response = file_response['entries'][0]
        return self.__class__(
            session=self._session,
            object_id=self._object_id,
            response_object=file_response,
        )

    @api_call
    def update_contents(
            self,
            file_path,
            etag=None,
            preflight_check=False,
            preflight_expected_size=0,
            upload_using_accelerator=False,
    ):
        """Upload a new version of a file. The contents are taken from the given file path.

        :param file_path:
            The path of the file that should be uploaded.
        :type file_path:
            `unicode`
        :param etag:
            If specified, instruct the Box API to update the item only if the current version's etag matches.
        :type etag:
            `unicode` or None
        :param preflight_check:
            If specified, preflight check will be performed before actually uploading the file.
        :type preflight_check:
            `bool`
        :param preflight_expected_size:
            The size of the file to be uploaded in bytes, which is used for preflight check. The default value is '0',
            which means the file size is unknown.
        :type preflight_expected_size:
            `int`
        :param upload_using_accelerator:
            If specified, the upload will try to use Box Accelerator to speed up the uploads for big files.
            It will make an extra API call before the actual upload to get the Accelerator upload url, and then make
            a POST request to that url instead of the default Box upload url. It falls back to normal upload endpoint,
            if cannot get the Accelerator upload url.

            Please notice that this is a premium feature, which might not be available to your app.
        :type upload_using_accelerator:
            `bool`
        :returns:
            A new file object
        :rtype:
            :class:`File`
        :raises:
            :class:`BoxAPIException` if the specified etag doesn't match the latest version of the file or preflight
            check fails.
        """
        with open(file_path, 'rb') as file_stream:
            return self.update_contents_with_stream(
                file_stream,
                etag,
                preflight_check,
                preflight_expected_size=preflight_expected_size,
                upload_using_accelerator=upload_using_accelerator,
            )

    @api_call
    def lock(self, prevent_download=False):
        """
        Lock a file, preventing others from modifying (or possibly even downloading) it.

        :param prevent_download:
            Whether or not the lock should prevent other users from downloading the file.
        :type prevent_download:
            `bool`
        :return:
            A new :class:`File` instance reflecting that the file has been locked.
        :rtype:
            :class:`File`
        """
        data = {
            'lock': {
                'is_download_prevented': prevent_download,
                'type': 'lock',
            }
        }
        return self.update_info(data)

    @api_call
    def unlock(self):
        """
        Unlock a file, releasing any restrictions that the lock maintained.

        :return:
            A new :class:`File` instance reflecting that the file has been unlocked.
        :rtype:
            :class:`File`
        """
        data = {'lock': None}
        return self.update_info(data)

    @api_call
    def get_shared_link_download_url(
            self,
            access=None,
            etag=None,
            unshared_at=None,
            allow_preview=None,
            password=None,
    ):
        """
        Get a shared link download url for the file with the given access permissions.
        This url is a direct download url for the file.

        :param access:
            Determines who can access the shared link. May be open, company, or collaborators. If no access is
            specified, the default access will be used.
        :type access:
            `unicode` or None
        :param etag:
            If specified, instruct the Box API to create the link only if the current version's etag matches.
        :type etag:
            `unicode` or None
        :param unshared_at:
            The date on which this link should be disabled. May only be set if the current user is not a free user
            and has permission to set expiration dates.
        :type unshared_at:
            :class:`datetime.date` or None
        :param allow_preview:
            Whether or not the item being shared can be previewed when accessed via the shared link.
            If this parameter is None, the default setting will be used.
        :type allow_preview:
            `bool` or None
        :param password:
            The password required to view this link. If no password is specified then no password will be set.
            Please notice that this is a premium feature, which might not be available to your app.
        :type password:
            `unicode` or None
        :returns:
            The URL of the shared link that allows direct download.
        :rtype:
            `unicode`
        :raises: :class:`BoxAPIException` if the specified etag doesn't match the latest version of the item.
        """
        item = self.create_shared_link(
            access=access,
            etag=etag,
            unshared_at=unshared_at,
            allow_preview=allow_preview,
            password=password,
        )
        return item.shared_link['download_url']



# coding: utf-8

from __future__ import absolute_import, unicode_literals

from mock import Mock
import pytest
from requests import Response

from boxsdk import Client
from boxsdk.auth.oauth2 import OAuth2
from test.integration.mock_network import MockNetwork


@pytest.fixture()
def box_client(box_oauth, mock_box_network):
    # pylint:disable=redefined-outer-name
    return Client(box_oauth, network_layer=mock_box_network)


@pytest.fixture()
def box_oauth(mock_box_network, client_id, client_secret, access_token, refresh_token):
    # pylint:disable=redefined-outer-name
    return OAuth2(
        client_id,
        client_secret,
        network_layer=mock_box_network,
        access_token=access_token,
        refresh_token=refresh_token,
    )


@pytest.fixture()
def mock_box_network():
    return MockNetwork()


@pytest.fixture
def generic_successful_response(generic_successful_request_response):
    return generic_successful_request_response


@pytest.fixture
def successful_token_mock(successful_token_request_response):
    return successful_token_request_response


@pytest.fixture(scope='session')
def unauthorized_response():
    res = Mock(Response)
    res.content = b''
    res.status_code = 401
    res.ok = False
    return res



# coding: utf-8

from __future__ import unicode_literals, absolute_import

from mock import Mock, PropertyMock
import pytest
from six.moves import range  # pylint:disable=redefined-builtin

from boxsdk.object.file import File
from boxsdk.object.folder import Folder
from boxsdk.pagination.page import Page
from boxsdk.session.box_session import BoxSession
from boxsdk.util.translator import Translator


@pytest.fixture()
def translator():
    return Translator()


@pytest.fixture()
def mock_session(translator):
    mock_box_session = Mock(BoxSession)
    type(mock_box_session).translator = PropertyMock(
        return_value=translator
    )
    return mock_box_session


@pytest.fixture()
def page_builder(mock_session):
    def factory_function(response):
        return Page(
            session=mock_session,
            response_object=response
        )
    return factory_function


@pytest.fixture()
def item_checker(mock_session):
    def item_checker_function(item_type, response_object, item):
        assert item.id == response_object['id']
        assert isinstance(item, item_type)
        assert item.response_object == response_object
        assert item.session is mock_session

    return item_checker_function


def test_response_property(page_builder):
    response = {1: 10, 2: 20, 3: 30}
    page = page_builder(response)
    assert response == page.response_object


def test_getitem(page_builder, item_checker):
    entry_1 = {
        "type": "folder",
        "id": "192429928",
        "sequence_id": "1",
        "etag": "1",
        "name": "Stephen Curry Three Pointers"
    }
    entry_2 = {
        "type": "file",
        "id": "818853862",
        "sequence_id": "0",
        "etag": "0",
        "name": "Warriors.jpg"
    }
    response_entries = [entry_1, entry_2]
    response = {"entries": response_entries}
    page = page_builder(response)

    item_checker(Folder, entry_1, page[0])
    item_checker(File, entry_2, page[1])
    assert page[1] == page[-1]


def test_getitem_past_length_raises(page_builder):
    length = 7
    response_entries = [object() for _ in range(length)]
    response = {"entries": response_entries}
    page = page_builder(response=response)

    with pytest.raises(IndexError):
        page[length]  # pylint:disable=pointless-statement


@pytest.mark.parametrize('length', (0, 1, 10))
def test_len(page_builder, length):
    response_entries = [object() for _ in range(length)]
    response = {
        "entries": response_entries
    }
    page = page_builder(response=response)
    assert len(page) == length


def test_translation_of_page_entries(page_builder, mock_session, item_checker):
    tested_item_types = ['folder', 'file', 'user', 'collaboration', 'group', 'foobar']
    response_entries = [{'type': item_type_string, 'id': str(i)} for i, item_type_string in enumerate(tested_item_types)]
    response = {"entries": response_entries}
    page = page_builder(response=response)
    for i, item in enumerate(page):
        item_checker(mock_session.translator.translate(response_entries[i]['type']), response_entries[i], item)



# coding: utf-8

from __future__ import unicode_literals, absolute_import
from requests.exceptions import Timeout

from .base_endpoint import BaseEndpoint
from ..util.api_call_decorator import api_call
from ..util.compat import with_metaclass
from ..util.enum import ExtendableEnumMeta
from ..util.lru_cache import LRUCache
from ..util.text_enum import TextEnum


# pylint:disable=too-many-ancestors
class EventsStreamType(with_metaclass(ExtendableEnumMeta, TextEnum)):
    """An enum of all possible values of the `stream_type` parameter for user events.

    The value of the `stream_type` parameter determines the type of events
    returned by the endpoint.

    <https://box-content.readme.io/reference#events>
    """


class UserEventsStreamType(EventsStreamType):
    """An enum of all possible values of the `stream_type` parameter for user events.

    - ALL: Returns all user events.
    - CHANGES: Returns tree changes.
    - SYNC: Returns tree changes only for sync folders.

    <https://box-content.readme.io/reference#standard-user-events>
    """
    ALL = 'all'
    CHANGES = 'changes'
    SYNC = 'sync'


class EnterpriseEventsStreamType(EventsStreamType):
    """An enum of all possible values of the `stream_type` parameter for enterprise events.

    - ADMIN_LOGS: Retrieves up to a year's events for all users in the enterprise.

    NOTE: Requires Admin: These stream types will only work with an auth token
    from an enterprise admin account.

    <https://box-content.readme.io/reference#enterprise-events>
    """
    ADMIN_LOGS = 'admin_logs'
# pylint:enable=too-many-ancestors


class Events(BaseEndpoint):
    """Box API endpoint for subscribing to changes in a Box account."""

    def get_url(self, *args):
        """Base class override."""
        # pylint:disable=arguments-differ
        return super(Events, self).get_url('events', *args)

    @api_call
    def get_events(self, limit=100, stream_position=0, stream_type=UserEventsStreamType.ALL):
        """
        Get Box events from a given stream position for a given stream type.

        :param limit:
            Maximum number of events to return.
        :type limit:
            `int`
        :param stream_position:
            The location in the stream from which to start getting events. 0 is the beginning of time. 'now' will
            return no events and just current stream position.

            NOTE: Currently, 'now' is only valid for user events stream types. The request will fail if an
            enterprise events stream type is passed.
        :type stream_position:
            `unicode`
        :param stream_type:
            (optional) Which type of events to return.
            Defaults to `UserEventsStreamType.ALL`.
        :type stream_type:
            :enum:`EventsStreamType`
        :returns:
            Dictionary containing the next stream position along with a list of some number of events.
        :rtype:
            `dict`
        """
        url = self.get_url()
        params = {
            'limit': limit,
            'stream_position': stream_position,
            'stream_type': stream_type,
        }
        box_response = self._session.get(url, params=params)
        response = box_response.json().copy()
        if 'entries' in response:
            response['entries'] = [self.translator.translate(item['type'])(item) for item in response['entries']]
        return response

    @api_call
    def get_latest_stream_position(self, stream_type=UserEventsStreamType.ALL):
        """
        Get the latest stream position. The return value can be used with :meth:`get_events` or
        :meth:`generate_events_with_long_polling`.

        :param stream_type:
            (optional) Which events stream to query.
            Defaults to `UserEventsStreamType.ALL`.

            NOTE: Currently, the Box API requires this to be one of the user
            events stream types. The request will fail if an enterprise events
            stream type is passed.
        :type stream_type:
            :enum:`UserEventsStreamType`
        :returns:
            The latest stream position.
        :rtype:
            `unicode`
        """
        return self.get_events(limit=0, stream_position='now', stream_type=stream_type)['next_stream_position']

    def _get_all_events_since(self, stream_position, stream_type=UserEventsStreamType.ALL):
        """
        :param stream_type:
            (optional) Which type of events to return.
            Defaults to `UserEventsStreamType.ALL`.
        :type stream_type:
            :enum:`EventsStreamType`
        """
        next_stream_position = stream_position
        while True:
            events = self.get_events(stream_position=next_stream_position, limit=100, stream_type=stream_type)
            next_stream_position = events['next_stream_position']
            events = events['entries']
            if not events:
                return
            for event in events:
                yield event, next_stream_position
            if len(events) < 100:
                return

    @api_call
    def long_poll(self, options, stream_position):
        """
        Set up a long poll connection at the specified url.

        :param options:
            The long poll options which include a long pull url, retry timeout, etc.
        :type options:
            `dict`
        :param stream_position:
            The location in the stream from which to start getting events. 0 is the beginning of time.
            'now' will return no events and just current stream position.
        :type stream_position:
            `unicode`
        :returns:
            {"message": "new_change"}, which means there're new changes on Box or {"version": 1, "message": "reconnect"}
            if nothing happens on Box during the long poll.
        :rtype:
            `dict`
        """
        url = options['url']
        long_poll_response = self._session.get(
            url,
            timeout=options['retry_timeout'],
            params={'stream_position': stream_position}
        )
        return long_poll_response

    @api_call
    def generate_events_with_long_polling(self, stream_position=None, stream_type=UserEventsStreamType.ALL):
        """
        Subscribe to events from the given stream position.

        :param stream_position:
            The location in the stream from which to start getting events. 0 is the beginning of time. 'now' will
            return no events and just current stream position.
        :type stream_position:
            `unicode`
        :param stream_type:
            (optional) Which type of events to return.
            Defaults to `UserEventsStreamType.ALL`.

            NOTE: Currently, the Box API requires this to be one of the user
            events stream types. The request will fail if an enterprise events
            stream type is passed.
        :type stream_type:
            :enum:`UserEventsStreamType`
        :returns:
            Events corresponding to changes on Box in realtime, as they come in.
        :rtype:
            `generator` of :class:`Event`
        """
        event_ids = LRUCache()
        stream_position = stream_position if stream_position is not None else self.get_latest_stream_position(stream_type=stream_type)
        while True:
            options = self.get_long_poll_options(stream_type=stream_type)
            while True:
                try:
                    long_poll_response = self.long_poll(options, stream_position)
                except Timeout:
                    break
                else:
                    message = long_poll_response.json()['message']
                    if message == 'new_change':
                        next_stream_position = stream_position
                        for event, next_stream_position in self._get_all_events_since(stream_position, stream_type=stream_type):
                            try:
                                event_ids.get(event['event_id'])
                            except KeyError:
                                yield event
                                event_ids.set(event['event_id'])
                        stream_position = next_stream_position
                        break
                    elif message == 'reconnect':
                        continue
                    else:
                        break

    @api_call
    def get_long_poll_options(self, stream_type=UserEventsStreamType.ALL):
        """
        Get the url and retry timeout for setting up a long polling connection.

        :param stream_type:
            (optional) Which type of events to return.
            Defaults to `UserEventsStreamType.ALL`.
        :type stream_type:
            :enum:`EventsStreamType`
        :returns:
            A `dict` including a long poll url, retry timeout, etc.
            E.g.
                {
                    "type": "realtime_server",
                    "url": "http://2.realtime.services.box.net/subscribe?channel=cc807c9c4869ffb1c81a&stream_type=all",
                    "ttl": "10",
                    "max_retries": "10",
                    "retry_timeout": 610,
                }
        :rtype:
            `dict`
        """
        url = self.get_url()
        params = {'stream_type': stream_type}
        box_response = self._session.options(url, params=params)
        return box_response.json()['entries'][0]



# coding: utf-8

from __future__ import unicode_literals

from boxsdk.object.recent_item import RecentItem


def test_init_recent_item(mock_box_session, mock_object_id):
    recent_item = RecentItem(
        session=mock_box_session,
        response_object={
            "type": "recent_item",
            "item": {"type": "file", "id": mock_object_id}
        })
    assert recent_item['type'] == 'recent_item'
    assert recent_item.item.object_id == mock_object_id
    assert recent_item.item.session is mock_box_session



# coding: utf-8

from __future__ import unicode_literals

from functools import partial
import re
from threading import Thread
import uuid

from mock import Mock, patch
import pytest
from six.moves import range   # pylint:disable=redefined-builtin
from six.moves.urllib import parse as urlparse  # pylint:disable=import-error,no-name-in-module,wrong-import-order

from boxsdk.exception import BoxOAuthException
from boxsdk.network.default_network import DefaultNetworkResponse
from boxsdk.auth.oauth2 import OAuth2
from boxsdk.config import API


class MyError(Exception):
    pass


class MyBaseException(BaseException):
    pass


@pytest.fixture(params=('https://url.com/foo?bar=baz', 'https://ȕŕľ.com/ƒőő?Ƅȁŕ=Ƅȁż', None))
def redirect_url(request):
    """A value for the `redirect_uri` query string parameter for OAuth2."""
    return request.param


def test_get_correct_authorization_url(redirect_url):
    # pylint:disable=redefined-outer-name
    fake_client_id = 'fake_client_id'
    fake_client_secret = 'fake_client_secret'
    oauth2 = OAuth2(
        client_id=fake_client_id,
        client_secret=fake_client_secret,
    )
    auth_url, csrf_token = oauth2.get_authorization_url(redirect_url=redirect_url)
    expected_auth_url_format = '{0}?state={1}&response_type=code&client_id={2}'
    if redirect_url:
        expected_auth_url_format += '&redirect_uri={3}'
    assert auth_url == expected_auth_url_format.format(
        API.OAUTH2_AUTHORIZE_URL,
        csrf_token,
        fake_client_id,
        urlparse.quote_plus((redirect_url or '').encode('utf-8')),
    )
    assert re.match('^box_csrf_token_[A-Za-z0-9]{16}$', csrf_token)


def test_authenticate_send_post_request_with_correct_params(mock_network_layer, successful_token_response):
    fake_client_id = 'fake_client_id'
    fake_client_secret = 'fake_client_secret'
    fake_auth_code = 'fake_auth_code'
    data = {
        'grant_type': 'authorization_code',
        'code': fake_auth_code,
        'client_id': fake_client_id,
        'client_secret': fake_client_secret,
        'box_device_id': '0',
        'box_device_name': 'my_awesome_device',
    }
    mock_network_layer.request.return_value = successful_token_response
    oauth = OAuth2(
        client_id=fake_client_id,
        client_secret=fake_client_secret,
        network_layer=mock_network_layer,
        box_device_name='my_awesome_device',
    )

    oauth.authenticate(fake_auth_code)

    mock_network_layer.request.assert_called_once_with(
        'POST',
        '{0}/token'.format(API.OAUTH2_API_URL),
        data=data,
        headers={'content-type': 'application/x-www-form-urlencoded'},
        access_token=None,
    )

    assert oauth.access_token == successful_token_response.json()['access_token']


@pytest.mark.parametrize('_', range(10))
def test_refresh_send_post_request_with_correct_params_and_handles_multiple_requests(
        mock_network_layer,
        successful_token_response,
        _,
):
    fake_client_id = 'fake_client_id'
    fake_client_secret = 'fake_client_secret'
    fake_refresh_token = 'fake_refresh_token'
    fake_access_token = 'fake_access_token'
    data = {
        'grant_type': 'refresh_token',
        'refresh_token': fake_refresh_token,
        'client_id': fake_client_id,
        'client_secret': fake_client_secret,
        'box_device_id': '0',
        'box_device_name': 'my_awesome_device',
    }
    mock_network_layer.request.return_value = successful_token_response
    oauth = OAuth2(
        client_id=fake_client_id,
        client_secret=fake_client_secret,
        access_token=fake_access_token,
        refresh_token=fake_refresh_token,
        network_layer=mock_network_layer,
        box_device_name='my_awesome_device',
    )

    # Create four threads to call refresh on oauth at the same time.
    threads = []
    for _ in range(4):
        threads.append(Thread(target=oauth.refresh, args=(fake_access_token,)))

    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()

    # Assert that even four threads were trying to refresh the tokens at the same time, only one token request was made,
    # and it was made with the correct params.
    mock_network_layer.request.assert_called_once_with(
        'POST',
        '{0}/token'.format(API.OAUTH2_API_URL),
        data=data,
        headers={'content-type': 'application/x-www-form-urlencoded'},
        access_token=fake_access_token,
    )


def test_authenticate_stores_tokens_correctly(mock_network_layer, successful_token_response):
    fake_client_id = 'fake_client_id'
    fake_client_secret = 'fake_client_secret'
    fake_auth_code = 'fake_auth_code'

    mock_network_layer.request.return_value = successful_token_response
    mock_token_callback = Mock()
    oauth = OAuth2(
        client_id=fake_client_id,
        client_secret=fake_client_secret,
        network_layer=mock_network_layer,
        store_tokens=mock_token_callback,
    )

    access_token, refresh_token = oauth.authenticate(fake_auth_code)
    mock_token_callback.assert_called_once_with(access_token, refresh_token)

    assert access_token == successful_token_response.json()['access_token']
    assert refresh_token == successful_token_response.json()['refresh_token']


@pytest.mark.parametrize('_', range(10))
def test_refresh_gives_back_the_correct_response_and_handles_multiple_requests(
        mock_network_layer,
        successful_token_response,
        network_response_with_missing_tokens,
        _,
):
    # pylint:disable=redefined-outer-name
    fake_client_id = 'fake_client_id'
    fake_client_secret = 'fake_client_secret'
    fake_refresh_token = 'fake_refresh_token'
    fake_access_token = 'fake_access_token'

    # Setup the network layer so that if oauth makes more than one request, it will get a malformed response and failed
    # the test.
    mock_network_layer.request.side_effect = [successful_token_response, network_response_with_missing_tokens]
    oauth = OAuth2(
        client_id=fake_client_id,
        client_secret=fake_client_secret,
        access_token=fake_access_token,
        refresh_token=fake_refresh_token,
        network_layer=mock_network_layer,
    )

    def refresh_tokens_and_verify_the_response():
        access_token, refresh_token = oauth.refresh(fake_access_token)
        assert access_token == successful_token_response.json()['access_token']
        assert refresh_token == successful_token_response.json()['refresh_token']

    # Creates four threads and do token refresh at the same time. Assert they all get the same new access token and
    # refresh token.
    threads = []
    for _ in range(4):
        threads.append(Thread(target=refresh_tokens_and_verify_the_response))

    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()


@pytest.fixture()
def token_method(request):
    """ Fixture that returns a partial method based on the method provided in request.param"""
    if request.param == OAuth2.refresh:
        return partial(OAuth2.refresh, access_token_to_refresh='fake_access_token')
    elif request.param == OAuth2.authenticate:
        return partial(OAuth2.authenticate, auth_code='fake_code')


@pytest.mark.parametrize(
    'token_method',
    [OAuth2.refresh, OAuth2.authenticate],
    indirect=True,
)
def test_token_request_raises_box_oauth_exception_when_getting_bad_network_response(
        token_method,
        mock_network_layer,
        bad_network_response,
):
    with pytest.raises(BoxOAuthException):
        mock_network_layer.request.return_value = bad_network_response
        oauth = OAuth2(
            client_id='',
            client_secret='',
            access_token='fake_access_token',
            network_layer=mock_network_layer,
        )
        token_method(oauth)


@pytest.mark.parametrize(
    'token_method',
    [OAuth2.refresh, OAuth2.authenticate],
    indirect=True,
)
def test_token_request_raises_box_oauth_exception_when_no_json_object_can_be_decoded(
        token_method,
        mock_network_layer,
        non_json_response,
):
    mock_network_layer.request.return_value = non_json_response
    oauth = OAuth2(
        client_id='',
        client_secret='',
        access_token='fake_access_token',
        network_layer=mock_network_layer,
    )
    with pytest.raises(BoxOAuthException):
        token_method(oauth)


@pytest.fixture(params=[
    ['access_token'],
    ['refresh_token'],
    [],
])
def network_response_with_missing_tokens(request):
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.ok = True
    json_dict = {}
    for key in request.param:
        json_dict[key] = 'fake_token'
    mock_network_response.json.return_value = json_dict
    return mock_network_response


@pytest.mark.parametrize('test_method', [
    partial(OAuth2.refresh, access_token_to_refresh='fake_access_token'),
    partial(OAuth2.authenticate, auth_code='fake_code')
])
def test_token_request_raises_box_oauth_exception_when_tokens_are_not_in_the_response(
        test_method,
        mock_network_layer,
        network_response_with_missing_tokens,
):
    # pylint:disable=redefined-outer-name
    mock_network_layer.request.return_value = network_response_with_missing_tokens
    oauth = OAuth2(
        client_id='',
        client_secret='',
        access_token='fake_access_token',
        network_layer=mock_network_layer,
    )
    with pytest.raises(BoxOAuthException):
        test_method(oauth)


def test_token_request_allows_missing_refresh_token(mock_network_layer):
    mock_network_response = Mock()
    mock_network_response.ok = True
    mock_network_response.json.return_value = {'access_token': 'fake_token'}
    mock_network_layer.request.return_value = mock_network_response
    oauth = OAuth2(
        client_id='',
        client_secret='',
        access_token='fake_access_token',
        network_layer=mock_network_layer,
    )
    oauth.send_token_request({}, access_token=None, expect_refresh_token=False)


@pytest.fixture()
def oauth(client_id, client_secret, access_token, refresh_token, mock_network_layer):
    return OAuth2(
        client_id=client_id,
        client_secret=client_secret,
        access_token=access_token,
        refresh_token=refresh_token,
        network_layer=mock_network_layer,
    )


@pytest.mark.parametrize(
    'access_token,refresh_token,expected_token_to_revoke',
    (
        ('fake_access_token', 'fake_refresh_token', 'fake_access_token'),
        (None, 'fake_refresh_token', 'fake_refresh_token')
    )
)
def test_revoke_sends_revoke_request(
        client_id,
        client_secret,
        mock_network_layer,
        access_token,
        oauth,
        expected_token_to_revoke,
):
    mock_network_response = Mock()
    mock_network_response.ok = True
    mock_network_layer.request.return_value = mock_network_response
    oauth.revoke()
    mock_network_layer.request.assert_called_once_with(
        'POST',
        '{0}/revoke'.format(API.OAUTH2_API_URL),
        data={
            'client_id': client_id,
            'client_secret': client_secret,
            'token': expected_token_to_revoke,
        },
        access_token=access_token,
    )
    assert oauth.access_token is None


def test_tokens_get_updated_after_noop_refresh(client_id, client_secret, access_token, new_access_token, refresh_token, mock_network_layer):
    """`OAuth2` object should update its state with new tokens, after no-op refresh.

    If the protected method `_get_tokens()` returns new tokens, refresh is
    skipped, and those tokens are used.

    This is a regression test for issue #128 [1]. We would return the new
    tokens without updating the object state. Subsequent uses of the `OAuth2`
    object would use the old tokens.

    [1] <https://github.com/box/box-python-sdk/issues/128>
    """
    new_refresh_token = uuid.uuid4().hex
    new_tokens = (new_access_token, new_refresh_token)

    class GetTokensOAuth2(OAuth2):
        def _get_tokens(self):
            """Return a new set of tokens, without updating any state.

            In order for the test to pass, the `OAuth2` object must be
            correctly programmed to take this return value and use it to update
            its state.
            """
            return new_tokens

    oauth = GetTokensOAuth2(
        client_id=client_id,
        client_secret=client_secret,
        access_token=access_token,
        refresh_token=refresh_token,
        network_layer=mock_network_layer,
    )
    assert oauth.access_token == access_token

    assert oauth.refresh(access_token) == new_tokens
    assert oauth.access_token == new_access_token


def test_closed_is_false_after_init(client_id, client_secret, mock_network_layer):
    auth = OAuth2(client_id=client_id, client_secret=client_secret, network_layer=mock_network_layer)
    assert auth.closed is False


def test_closed_is_true_after_close(client_id, client_secret, mock_network_layer):
    auth = OAuth2(client_id=client_id, client_secret=client_secret, network_layer=mock_network_layer)
    auth.close()
    assert auth.closed is True


def test_token_requests_fail_after_close(client_id, client_secret, mock_network_layer):
    auth = OAuth2(client_id=client_id, client_secret=client_secret, network_layer=mock_network_layer)
    auth.close()
    with pytest.raises(ValueError):
        auth.refresh(auth.access_token)


@pytest.mark.parametrize('raise_exception', [False, True])
def test_context_manager_closes_auth_object(client_id, client_secret, mock_network_layer, raise_exception):
    auth = OAuth2(client_id=client_id, client_secret=client_secret, network_layer=mock_network_layer)
    try:
        with auth.closing():
            if raise_exception:
                raise MyError
    except MyError:
        pass
    assert auth.closed is True


def test_context_manager_fails_after_close(client_id, client_secret, mock_network_layer):
    auth = OAuth2(client_id=client_id, client_secret=client_secret, network_layer=mock_network_layer)
    with auth.closing():
        pass
    with pytest.raises(ValueError):
        with auth.closing():
            assert False


@pytest.mark.parametrize(('close_args', 'close_kwargs'), [((), {}), ((True,), {}), ((), dict(revoke=True))])
def test_revoke_on_close(client_id, client_secret, access_token, mock_network_layer, close_args, close_kwargs):
    auth = OAuth2(client_id=client_id, client_secret=client_secret, access_token=access_token, network_layer=mock_network_layer)
    with patch.object(auth, 'revoke') as mock_revoke:
        auth.close(*close_args, **close_kwargs)
    mock_revoke.assert_called_once_with()


def test_auth_object_is_closed_even_if_revoke_fails(client_id, client_secret, access_token, mock_network_layer):
    auth = OAuth2(client_id=client_id, client_secret=client_secret, access_token=access_token, network_layer=mock_network_layer)
    with patch.object(auth, 'revoke', side_effect=BoxOAuthException(status=500)):
        with pytest.raises(BoxOAuthException):
            auth.close(revoke=True)
    assert auth.closed is True


@pytest.mark.parametrize(('close_args', 'close_kwargs'), [((False,), {}), ((), dict(revoke=False))])
def test_revoke_on_close_can_be_skipped(client_id, client_secret, access_token, mock_network_layer, close_args, close_kwargs):
    auth = OAuth2(client_id=client_id, client_secret=client_secret, access_token=access_token, network_layer=mock_network_layer)
    with patch.object(auth, 'revoke') as mock_revoke:
        auth.close(*close_args, **close_kwargs)
    mock_revoke.assert_not_called()


@pytest.mark.parametrize(('raise_from_block', 'raise_from_close', 'expected_exception'), [
    (MyError, None, MyError),
    (None, BoxOAuthException(status=500), BoxOAuthException),
    (MyError, BoxOAuthException(status=500), MyError),
])
@pytest.mark.parametrize('close_kwargs', [{}, dict(revoke=False), dict(revoke=True)])
def test_context_manager_reraises_first_exception_after_close(
        client_id, client_secret, mock_network_layer, close_kwargs, raise_from_block, raise_from_close, expected_exception,
):
    auth = OAuth2(client_id=client_id, client_secret=client_secret, network_layer=mock_network_layer)
    with patch.object(auth, 'close', side_effect=raise_from_close) as mock_close:
        with pytest.raises(expected_exception):
            with auth.closing(**close_kwargs):
                if raise_from_block:
                    raise raise_from_block
    mock_close.assert_called_once_with(**close_kwargs)


@pytest.mark.parametrize('close_kwargs', [{}, dict(revoke=False), dict(revoke=True)])
def test_context_manager_skips_revoke_on_base_exception(client_id, client_secret, mock_network_layer, close_kwargs):
    auth = OAuth2(client_id=client_id, client_secret=client_secret, network_layer=mock_network_layer)
    with patch.object(auth, 'close') as mock_close:
        with pytest.raises(MyBaseException):
            with auth.closing(**close_kwargs):
                raise MyBaseException
    mock_close.assert_called_once_with(revoke=False)



# coding: utf-8

from __future__ import unicode_literals, absolute_import


class Cloneable(object):
    """
    Cloneable interface to be implemented by endpoint objects that should have ability to be cloned, but with a
    different session member if desired.
    """

    def as_user(self, user):
        """
        Returns a new endpoint object with default headers set up to make requests as the specified user.

        :param user:
            The user to impersonate when making API requests.
        :type user:
            :class:`User`
        """
        return self.clone(self.session.as_user(user))

    def with_shared_link(self, shared_link, shared_link_password):
        """
        Returns a new endpoint object with default headers set up to make requests using the shared link for auth.

        :param shared_link:
            The shared link.
        :type shared_link:
            `unicode`
        :param shared_link_password:
            The password for the shared link.
        :type shared_link_password:
            `unicode`
        """
        return self.clone(self.session.with_shared_link(shared_link, shared_link_password))

    def clone(self, session=None):
        """
        Returns a copy of this cloneable object using the specified session.

        :param session:
            The Box session used to make requests.
        :type session:
            :class:`BoxSession`
        """
        raise NotImplementedError

    @property
    def session(self):
        """
        Return the Box session being used to make requests.

        :rtype:
            :class:`BoxSession`
        """
        raise NotImplementedError



# coding: utf-8

from __future__ import unicode_literals, absolute_import

from .developer_token_client import DeveloperTokenClient
from .logging_client import LoggingClient


class DevelopmentClient(DeveloperTokenClient, LoggingClient):
    """
    Client subclass that uses developer token auth and logs requests and responses.
    Great for use in development!
    """
    pass



# coding: utf-8

from __future__ import absolute_import, division, unicode_literals

from datetime import timedelta

import six
from six.moves import map


NoneType = type(None)


if not hasattr(timedelta, 'total_seconds'):
    def total_seconds(delta):
        """
        Return the total number of seconds represented by this timedelta.
        Python 2.6 does not define this method.
        """
        return (delta.microseconds + (delta.seconds + delta.days * 24 * 3600) * 10 ** 6) / 10 ** 6
else:
    def total_seconds(delta):
        return delta.total_seconds()


def with_metaclass(meta, *bases, **with_metaclass_kwargs):
    """Extends the behavior of six.with_metaclass.

    The normal usage (expanded to include temporaries, to make the illustration
    easier) is:

    .. code-block:: python

        temporary_class = six.with_metaclass(meta, *bases)
        temporary_metaclass = type(temporary_class)

        class Subclass(temporary_class):
            ...

        SubclassMeta = type(Subclass)

    In this example:

    - ``temporary_class`` is a class with ``(object,)`` as its bases.
    - ``temporary_metaclass`` is a metaclass with ``(meta,)`` as its bases.
    - ``Subclass`` is a class with ``bases`` as its bases.
    - ``SubclassMeta`` is ``meta``.

    ``six.with_metaclass()`` is defined in such a way that it can make sure
    that ``Subclass`` has the correct metaclass and bases, while only using
    syntax which is common to both Python 2 and Python 3.
    ``temporary_metaclass()`` returns an instance of ``meta``, rather than an
    instance of itself / a subclass of ``temporary_class``, which is how
    ``SubclassMeta`` ends up being ``meta``, and how the temporaries don't
    appear anywhere in the final subclass.

    There are two problems with the current (as of six==1.10.0) implementation
    of ``six.with_metaclass()``, which this function solves.

    ``six.with_metaclass()`` does not define ``__prepare__()`` on the temporary
    metaclass. This means that ``meta.__prepare__()`` gets called directly,
    with bases set to ``(object,)``. If it needed to actually receive
    ``bases``, then errors might occur. For example, this was a problem when
    used with ``enum.EnumMeta`` in Python 3.6. Here we make sure that
    ``__prepare__()`` is defined on the temporary metaclass, and pass ``bases``
    to ``meta.__prepare__()``. This is fixed in six>=1.11.0 by PR #178 [1].

    Since ``temporary_class`` doesn't have the correct bases, in theory this
    could cause other problems, besides the previous one, in certain edge
    cases. To make sure that doesn't become a problem, we make sure that
    ``temporary_class`` has ``bases`` as its bases, just like the final class.

    [1] <https://github.com/benjaminp/six/pull/178>
    """
    temporary_class = six.with_metaclass(meta, *bases, **with_metaclass_kwargs)
    temporary_metaclass = type(temporary_class)

    class TemporaryMetaSubclass(temporary_metaclass, _most_derived_metaclass(meta, bases)):

        if '__prepare__' not in temporary_metaclass.__dict__:
            # six<1.11.0, __prepare__ is not defined on the temporary metaclass.

            @classmethod
            def __prepare__(mcs, name, this_bases, **kwds):  # pylint:disable=unused-argument,arguments-differ
                return meta.__prepare__(name, bases, **kwds)

    return type.__new__(TemporaryMetaSubclass, str('temporary_class'), bases, {})


def _most_derived_metaclass(meta, bases):
    """Selects the most derived metaclass of all the given metaclasses.

    This will be the same metaclass that is selected by

    .. code-block:: python

        class temporary_class(*bases, metaclass=meta): pass

    or equivalently by

    .. code-block:: python

        types.prepare_class('temporary_class', bases, metaclass=meta)

    "Most derived" means the item in {meta, type(bases[0]), type(bases[1]), ...}
    which is a non-strict subclass of every item in that set.

    If no such item exists, then :exc:`TypeError` is raised.

    :type meta:   `type`
    :type bases:  :class:`Iterable` of `type`
    """
    most_derived_metaclass = meta
    for base_type in map(type, bases):
        if issubclass(base_type, most_derived_metaclass):
            most_derived_metaclass = base_type
        elif not issubclass(most_derived_metaclass, base_type):
            # Raises TypeError('metaclass conflict: ...')
            return type.__new__(meta, str('temporary_class'), bases, {})
    return most_derived_metaclass



# coding: utf-8

from __future__ import unicode_literals

from .cooperatively_managed_oauth2 import CooperativelyManagedOAuth2
from .developer_token_auth import DeveloperTokenAuth
try:
    from .jwt_auth import JWTAuth
except ImportError:
    JWTAuth = None  # If extras[jwt] are not installed, JWTAuth won't be available.
from .oauth2 import OAuth2
try:
    from .redis_managed_oauth2 import RedisManagedOAuth2
except ImportError:
    RedisManagedOAuth2 = None  # If extras[redis] are not installed, RedisManagedOAuth2 won't be available.
try:
    from .redis_managed_jwt_auth import RedisManagedJWTAuth
except ImportError:
    RedisManagedJWTAuth = None  # If extras[jwt,redis] are not installed, RedisManagedJWTAuth won't be available.
from .remote_managed_oauth2 import RemoteOAuth2



# coding: utf-8

from __future__ import unicode_literals
# pylint:disable=unused-import, import-error
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict



# coding: utf-8

from __future__ import unicode_literals
import pytest
from boxsdk.util.lru_cache import LRUCache


@pytest.fixture
def lru_cache():
    return LRUCache()


@pytest.fixture
def keys():
    return ['key1', 'key2', 'key3']


def test_lru_cache_returns_minus_one_for_missing_key(lru_cache, keys):
    # pylint:disable=redefined-outer-name
    for key in keys:
        with pytest.raises(KeyError):
            lru_cache.get(key)


def test_lru_cache_returns_none_for_existing_key(lru_cache, keys):
    # pylint:disable=redefined-outer-name
    for key in keys:
        lru_cache.set(key)
        assert lru_cache.get(key) is None


def test_lru_cache_ejects_least_recently_used_key(lru_cache, keys):
    # pylint:disable=redefined-outer-name
    lru_cache.capacity = len(keys)
    for key in keys:
        lru_cache.set(key)
    lru_cache.set('another key')
    with pytest.raises(KeyError):
        lru_cache.get(keys[0])
    assert lru_cache.get('another key') is None
    for key in keys[1:]:
        assert lru_cache.get(key) is None
    lru_cache.set('yet another key')
    with pytest.raises(KeyError):
        lru_cache.get('another key')



# coding: utf-8

from __future__ import absolute_import, unicode_literals
from enum import Enum
from six import text_type


class TextEnum(text_type, Enum):
    def __repr__(self):
        return self._value_  # pylint:disable=no-member

    def __str__(self):
        return str(self.value)  # pylint:disable=no-member



# coding: utf-8

from __future__ import unicode_literals
import pytest

from boxsdk.object.group_membership import GroupMembership
from boxsdk.object.group import Group
from boxsdk.object.user import User


@pytest.fixture(params=[True, False])
def mock_group_membership_dict_or_none(request, mock_group_membership_dict):
    return mock_group_membership_dict if request.param else None


@pytest.fixture(params=[True, False])
def mock_user_and_group_or_none(request, mock_user, mock_group):
    return (mock_user, mock_group) if request.param else None


def test_group_membership_initialization(
        mock_box_session,
        mock_group_membership_dict_or_none,
        mock_user_and_group_or_none,
):
    # pylint:disable=redefined-outer-name
    fake_id = "fake_membership_id"
    response_object = mock_group_membership_dict_or_none

    kwargs = {}

    if mock_user_and_group_or_none:
        kwargs['user'], kwargs['group'] = mock_user_and_group_or_none
        user_id, group_id = kwargs['user'].object_id, kwargs['group'].object_id
    else:
        user_id, group_id = None, None

    group_membership = GroupMembership(mock_box_session, fake_id, response_object, **kwargs)

    assert group_membership.object_id == fake_id

    # pylint:disable=protected-access
    assert group_membership._session == mock_box_session
    if response_object:
        assert group_membership._response_object is response_object
    else:
        assert not group_membership._response_object
    # pylint:enable=protected-access

    _assert_object_is_valid(group_membership.user, bool(response_object), user_id, User, kwargs.get('user'))
    _assert_object_is_valid(group_membership.group, bool(response_object), group_id, Group, kwargs.get('group'))


def _assert_object_is_valid(actual_object, has_response, object_id, expected_class, expected_object):
    if has_response or object_id:
        assert isinstance(actual_object, expected_class)
    else:
        assert actual_object is None
    if object_id:
        assert actual_object is expected_object



# coding: utf-8

from __future__ import unicode_literals
from .oauth2 import OAuth2


class CooperativelyManagedOAuth2Mixin(OAuth2):
    """
    Box SDK OAuth2 mixin.
    Allows for sharing auth tokens between multiple clients.
    """
    def __init__(self, retrieve_tokens=None, *args, **kwargs):
        """
        :param retrieve_tokens:
            Callback to get the current access/refresh token pair.
        :type retrieve_tokens:
            `callable` of () => (`unicode`, `unicode`)
        """
        self._retrieve_tokens = retrieve_tokens
        super(CooperativelyManagedOAuth2Mixin, self).__init__(*args, **kwargs)

    def _get_tokens(self):
        """
        Base class override. Get the tokens from the user-specified callback.
        """
        return self._retrieve_tokens()


class CooperativelyManagedOAuth2(CooperativelyManagedOAuth2Mixin):
    """
    Box SDK OAuth2 subclass.
    Allows for sharing auth tokens between multiple clients. The retrieve_tokens callback should
    return the current access/refresh token pair.
    """
    pass



# coding: utf-8

from __future__ import unicode_literals, absolute_import

from .auth import JWTAuth, OAuth2
from .client import *  # pylint:disable=wildcard-import,redefined-builtin
from .object import *  # pylint:disable=wildcard-import,redefined-builtin
from .version import __version__



# coding: utf-8

from __future__ import unicode_literals, absolute_import

from .client import Client
from ..network.logging_network import LoggingNetwork


class LoggingClient(Client):
    """
    Box client subclass which logs requests and responses.
    """
    def __init__(self, oauth=None, network_layer=None, session=None):
        super(LoggingClient, self).__init__(
            oauth=oauth,
            network_layer=network_layer or LoggingNetwork(),
            session=session,
        )



# coding: utf-8

from __future__ import unicode_literals
import logging
import sys

from six import string_types


def setup_logging(stream_or_file=None, debug=False, name=None):
    """
    Create a logger for communicating with the user or writing to log files.
    By default, creates a root logger that prints to stdout.

    :param stream_or_file:
        The destination of the log messages. If None, stdout will be used.
    :type stream_or_file:
        `unicode` or `file` or None
    :param debug:
        Whether or not the logger will be at the DEBUG level (if False, the logger will be at the INFO level).
    :type debug:
        `bool` or None
    :param name:
        The logging channel. If None, a root logger will be created.
    :type name:
        `unicode` or None
    :return:
        A logger that's been set up according to the specified parameters.
    :rtype:
        :class:`Logger`
    """
    logger = logging.getLogger(name)
    if isinstance(stream_or_file, string_types):
        logger.addHandler(logging.FileHandler(stream_or_file, mode='w'))
    else:
        logger.addHandler(logging.StreamHandler(stream_or_file or sys.stdout))
    logger.setLevel(logging.DEBUG if debug else logging.INFO)
    return logger



# coding: utf-8

from __future__ import absolute_import, unicode_literals

from itertools import product

import pytest

from boxsdk.object.base_object import BaseObject
from boxsdk.object.file import File
from boxsdk.object.folder import Folder
from boxsdk.object.group import Group
from boxsdk.object.user import User
from boxsdk.util.translator import Translator


_response_to_class_mapping = {}


@pytest.fixture
def bookmark_response(make_mock_box_request, mock_object_id):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={'type': 'bookmark', 'id': mock_object_id},
    )
    return mock_box_response


@pytest.fixture
def box_note_response(make_mock_box_request, mock_object_id):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={'type': 'boxnote', 'id': mock_object_id},
    )
    return mock_box_response


@pytest.fixture(autouse=True)
def translator_response(
        bookmark_response,
        box_note_response,
        mock_file_response,
        mock_folder_response,
        mock_group_response,
        mock_user_response,
):
    # pylint:disable=redefined-outer-name
    _response_to_class_mapping['bookmark'] = (bookmark_response, BaseObject)
    _response_to_class_mapping['box_note'] = (box_note_response, BaseObject)
    _response_to_class_mapping['file'] = (mock_file_response, File)
    _response_to_class_mapping['folder'] = (mock_folder_response, Folder)
    _response_to_class_mapping['group'] = (mock_group_response, Group)
    _response_to_class_mapping['user'] = (mock_user_response, User)


@pytest.mark.parametrize('response_type', ['bookmark', 'box_note', 'file', 'folder', 'group', 'user'])
def test_translator_converts_response_to_correct_type(response_type):
    response, object_class = _response_to_class_mapping[response_type]
    assert type(Translator().translate(response.json()['type']) == object_class)


def test_default_translator():
    assert isinstance(Translator._default_translator, Translator)   # pylint:disable=protected-access


@pytest.mark.parametrize(('extend_default_translator', 'new_child'), list(product([None, True], [False, True])))
def test_with_extend_default_translator(default_translator, extend_default_translator, new_child):
    item_type = 'foo'

    class Foo(object):
        pass

    kwargs = {}
    if extend_default_translator is not None:
        kwargs['extend_default_translator'] = extend_default_translator
    translator = Translator({item_type: Foo}, new_child=new_child, **kwargs)
    assert set(translator.items()).issuperset(default_translator.items())


@pytest.mark.parametrize('new_child', [False, True])
def test_without_extend_default_translator(new_child):
    item_type = 'foo'

    class Foo(object):
        pass

    mapping = {item_type: Foo}

    translator = Translator(mapping, extend_default_translator=False, new_child=new_child)
    assert translator == mapping


@pytest.mark.parametrize(('new_child', 'extend_default_translator'), list(product([None, True], [False, True])))
def test_with_new_child(new_child, extend_default_translator):
    item_type = 'foo'

    class Foo(object):
        pass

    mapping = {item_type: Foo}

    kwargs = {}
    if new_child is not None:
        kwargs['new_child'] = new_child

    translator = Translator(mapping, extend_default_translator=extend_default_translator, **kwargs)
    assert item_type in translator
    assert translator.maps[0] == {}
    with pytest.raises(KeyError):
        del translator[item_type]

    class Bar(Foo):
        pass

    translator.register(item_type, Bar)
    assert translator.translate(item_type) is Bar
    assert mapping == {item_type: Foo}


@pytest.mark.parametrize('extend_default_translator', [False, True])
def test_without_new_child(extend_default_translator):
    item_type = 'foo'

    class Foo(object):
        pass

    mapping = {item_type: Foo}

    translator = Translator(mapping, new_child=False, extend_default_translator=extend_default_translator)
    assert item_type in translator
    assert translator.maps[0] is mapping

    class Bar(Foo):
        pass

    translator.register(item_type, Bar)
    assert translator.translate(item_type) is Bar
    assert mapping == {item_type: Bar}

    del translator[item_type]
    assert not mapping



# coding: utf-8

from __future__ import unicode_literals, absolute_import

from collections import Sequence
import copy

from boxsdk.object.base_object import BaseObject
from boxsdk.object.base_endpoint import BaseEndpoint


class Page(Sequence, object):
    """
    A sequence of BaseObjects that belong to a page returned from a paging api call.

    The Page makes available detailed response data for page requests.
    """
    _item_entries_key_name = "entries"

    def __init__(self, session, response_object):
        """
        :param session:
            The Box session used to make the request that generated the response.
        :type session:
            :class:`BoxSession`
        :param response_object:
            The parsed HTTP response from Box after requesting more pages.
        :type response_object:
            `dict`
        """
        super(Page, self).__init__()
        self._session = session
        self._response_object = response_object

    @property
    def _translator(self):
        """The translator used for translating Box API JSON responses into `BaseAPIJSONObject` smart objects.

        :rtype:   :class:`Translator`
        """
        return self._session.translator

    @property
    def response_object(self):
        """
        Return a copy of the response object for this Page.

        :rtype: `dict`
        """
        return copy.deepcopy(self._response_object)

    def __getitem__(self, key):
        """
        Try to get the attribute from the API response object.

        :param key:
            The attribute to retrieve from the API response object.
        :type key:
            `unicode`
        :rtype:
            :class:`BaseObject`
        """
        item_json = self._response_object[self._item_entries_key_name][key]
        item_class = self._translator.translate(item_json['type'])
        kwargs = {}
        if issubclass(item_class, BaseObject):
            kwargs['object_id'] = item_json['id']
        if issubclass(item_class, BaseEndpoint):
            kwargs['session'] = self._session

        item = item_class(response_object=item_json, **kwargs)
        return item

    def __len__(self):
        """
        Get the number of items in the page.

        :rtype:
            `int`
        """
        return len(self._response_object[self._item_entries_key_name])



# coding: utf-8

from __future__ import unicode_literals, absolute_import

from mock import Mock, patch
from boxsdk.auth import developer_token_auth


def test_developer_token_auth_calls_callback_during_init_and_refresh(access_token):
    get_new_token_callback = Mock()
    get_new_token_callback.return_value = access_token
    auth = developer_token_auth.DeveloperTokenAuth(
        get_new_token_callback=get_new_token_callback,
    )
    assert auth.access_token == access_token
    get_new_token_callback.assert_called_once_with()
    assert auth.refresh(access_token) == (access_token, None)
    assert len(get_new_token_callback.mock_calls) == 2


def test_developer_token_auth_uses_raw_input_by_default(access_token):
    with patch('boxsdk.auth.developer_token_auth.input', create=True) as mock_raw_input:
        mock_raw_input.return_value = access_token
        auth = developer_token_auth.DeveloperTokenAuth()
        mock_raw_input.assert_called_once_with(auth.ENTER_TOKEN_PROMPT)
        assert auth.access_token == access_token
        assert auth.refresh(access_token) == (access_token, None)
        assert len(mock_raw_input.mock_calls) == 2



# coding: utf-8

from __future__ import unicode_literals, absolute_import

from .base_endpoint import BaseEndpoint
from .base_api_json_object import BaseAPIJSONObject


class RecentItem(BaseEndpoint, BaseAPIJSONObject):
    """Represents a single recent item accessed by a Box user."""

    _item_type = 'recent_item'

    @property
    def item(self):
        """
        Returns the Box Item which this recent item references.

        :rtype:
            :class:`Item`
        """
        item = self._response_object['item']
        return self.translator.translate(item['type'])(
            session=self._session,
            object_id=item['id'],
            response_object=item,
        )



# coding: utf-8

from __future__ import unicode_literals, absolute_import

from mock import Mock

from boxsdk.auth import cooperatively_managed_oauth2


def test_cooperatively_managed_oauth2_calls_retrieve_tokens_during_refresh(access_token, refresh_token):
    retrieve_tokens = Mock()
    oauth2 = cooperatively_managed_oauth2.CooperativelyManagedOAuth2(
        retrieve_tokens=retrieve_tokens,
        client_id=None,
        client_secret=None,
    )
    retrieve_tokens.return_value = access_token, refresh_token
    assert oauth2.refresh(None) == (access_token, refresh_token)
    retrieve_tokens.assert_called_once_with()



# coding: utf-8

from __future__ import absolute_import, unicode_literals

from datetime import datetime, timedelta
import json
import random
import string

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
import jwt
from six import binary_type, string_types, raise_from, text_type

from .oauth2 import OAuth2
from ..object.user import User
from ..util.compat import NoneType, total_seconds


class JWTAuth(OAuth2):
    """
    Responsible for handling JWT Auth for Box Developer Edition. Can authenticate enterprise instances or app users.
    """
    _GRANT_TYPE = 'urn:ietf:params:oauth:grant-type:jwt-bearer'

    def __init__(
            self,
            client_id,
            client_secret,
            enterprise_id,
            jwt_key_id,
            rsa_private_key_file_sys_path=None,
            rsa_private_key_passphrase=None,
            user=None,
            store_tokens=None,
            box_device_id='0',
            box_device_name='',
            access_token=None,
            network_layer=None,
            jwt_algorithm='RS256',
            rsa_private_key_data=None,
            **kwargs
    ):
        """Extends baseclass method.

        Must pass exactly one of either `rsa_private_key_file_sys_path` or
        `rsa_private_key_data`.

        If both `enterprise_id` and `user` are non-`None`, the `user` takes
        precedence when `refresh()` is called. This can be overruled with a
        call to `authenticate_instance()`.

        :param client_id:
            Box API key used for identifying the application the user is authenticating with.
        :type client_id:
            `unicode`
        :param client_secret:
            Box API secret used for making OAuth2 requests.
        :type client_secret:
            `unicode`
        :param enterprise_id:
            The ID of the Box Developer Edition enterprise.

            May be `None`, if the caller knows that it will not be
            authenticating as an enterprise instance / service account.

            If `user` is passed, this value is not used, unless
            `authenticate_instance()` is called to clear the user and
            authenticate as the enterprise instance.
        :type enterprise_id:
            `unicode` or `None`
        :param jwt_key_id:
            Key ID for the JWT assertion.
        :type jwt_key_id:
            `unicode`
        :param rsa_private_key_file_sys_path:
            (optional) Path to an RSA private key file, used for signing the JWT assertion.
        :type rsa_private_key_file_sys_path:
            `unicode`
        :param rsa_private_key_passphrase:
            Passphrase used to unlock the private key. Do not pass a unicode string - this must be bytes.
        :type rsa_private_key_passphrase:
            `bytes` or None
        :param user:
            (optional) The user to authenticate, expressed as a Box User ID or
            as a :class:`User` instance.

            This value is not required. But if it is provided, then the user
            will be auto-authenticated at the time of the first API call or
            when calling `authenticate_user()` without any arguments.

            Should be `None` if the intention is to authenticate as the
            enterprise instance / service account. If both `enterprise_id` and
            `user` are non-`None`, the `user` takes precedense when `refresh()`
            is called.

            May be one of this application's created App User. Depending on the
            configured User Access Level, may also be any other App User or
            Managed User in the enterprise.

            <https://docs.box.com/docs/configuring-box-platform#section-3-enabling-app-auth-and-app-users>
            <https://docs.box.com/docs/authentication#section-choosing-an-authentication-type>
        :type user:
            `unicode` or :class:`User` or `None`
        :param store_tokens:
            Optional callback for getting access to tokens for storing them.
        :type store_tokens:
            `callable`
        :param box_device_id:
            Optional unique ID of this device. Used for applications that want to support device-pinning.
        :type box_device_id:
            `unicode`
        :param box_device_name:
            Optional human readable name for this device.
        :type box_device_name:
            `unicode`
        :param access_token:
            Access token to use for auth until it expires.
        :type access_token:
            `unicode`
        :param network_layer:
            If specified, use it to make network requests. If not, the default network implementation will be used.
        :type network_layer:
            :class:`Network`
        :param jwt_algorithm:
            Which algorithm to use for signing the JWT assertion. Must be one of 'RS256', 'RS384', 'RS512'.
        :type jwt_algorithm:
            `unicode`
        :param rsa_private_key_data:
            (optional) Contents of RSA private key, used for signing the JWT assertion. Do not pass a
            unicode string. Can pass a byte string, or a file-like object that returns bytes, or an
            already-loaded `RSAPrivateKey` object.
        :type rsa_private_key_data:   `bytes` or :class:`io.IOBase` or :class:`RSAPrivateKey`
        """
        user_id = self._normalize_user_id(user)
        rsa_private_key = self._normalize_rsa_private_key(
            file_sys_path=rsa_private_key_file_sys_path,
            data=rsa_private_key_data,
            passphrase=rsa_private_key_passphrase,
        )
        del rsa_private_key_data
        del rsa_private_key_file_sys_path
        super(JWTAuth, self).__init__(
            client_id,
            client_secret,
            store_tokens=store_tokens,
            box_device_id=box_device_id,
            box_device_name=box_device_name,
            access_token=access_token,
            refresh_token=None,
            network_layer=network_layer,
            **kwargs
        )
        self._rsa_private_key = rsa_private_key
        self._enterprise_id = enterprise_id
        self._jwt_algorithm = jwt_algorithm
        self._jwt_key_id = jwt_key_id
        self._user_id = user_id

    def _auth_with_jwt(self, sub, sub_type):
        """
        Get an access token for use with Box Developer Edition. Pass an enterprise ID to get an enterprise token
        (which can be used to provision/deprovision users), or a user ID to get a user token.

        :param sub:
            The enterprise ID or user ID to auth.
        :type sub:
            `unicode`
        :param sub_type:
            Either 'enterprise' or 'user'
        :type sub_type:
            `unicode`
        :return:
            The access token for the enterprise or app user.
        :rtype:
            `unicode`
        """
        system_random = random.SystemRandom()
        jti_length = system_random.randint(16, 128)
        ascii_alphabet = string.ascii_letters + string.digits
        ascii_len = len(ascii_alphabet)
        jti = ''.join(ascii_alphabet[int(system_random.random() * ascii_len)] for _ in range(jti_length))
        now_plus_30 = datetime.utcnow() + timedelta(seconds=30)
        assertion = jwt.encode(
            {
                'iss': self._client_id,
                'sub': sub,
                'box_sub_type': sub_type,
                'aud': 'https://api.box.com/oauth2/token',
                'jti': jti,
                'exp': int(total_seconds(now_plus_30 - datetime(1970, 1, 1))),
            },
            self._rsa_private_key,
            algorithm=self._jwt_algorithm,
            headers={
                'kid': self._jwt_key_id,
            },
        )
        data = {
            'grant_type': self._GRANT_TYPE,
            'client_id': self._client_id,
            'client_secret': self._client_secret,
            'assertion': assertion,
        }
        if self._box_device_id:
            data['box_device_id'] = self._box_device_id
        if self._box_device_name:
            data['box_device_name'] = self._box_device_name
        return self.send_token_request(data, access_token=None, expect_refresh_token=False)[0]

    def authenticate_user(self, user=None):
        """
        Get an access token for a User.

        May be one of this application's created App User. Depending on the
        configured User Access Level, may also be any other App User or Managed
        User in the enterprise.

        <https://docs.box.com/docs/configuring-box-platform#section-3-enabling-app-auth-and-app-users>
        <https://docs.box.com/docs/authentication#section-choosing-an-authentication-type>

        :param user:
            (optional) The user to authenticate, expressed as a Box User ID or
            as a :class:`User` instance.

            If not given, then the most recently provided user ID, if
            available, will be used.
        :type user:
            `unicode` or :class:`User`
        :raises:
            :exc:`ValueError` if no user ID was passed and the object is not
            currently configured with one.
        :return:
            The access token for the user.
        :rtype:
            `unicode`
        """
        sub = self._normalize_user_id(user) or self._user_id
        if not sub:
            raise ValueError("authenticate_user: Requires the user ID, but it was not provided.")
        self._user_id = sub
        return self._auth_with_jwt(sub, 'user')

    authenticate_app_user = authenticate_user

    @classmethod
    def _normalize_user_id(cls, user):
        """Get a Box user ID from a selection of supported param types.

        :param user:
            An object representing the user or user ID.

            Currently supported types are `unicode` (which represents the user
            ID) and :class:`User`.

            If `None`, returns `None`.
        :raises:  :exc:`TypeError` for unsupported types.
        :rtype:   `unicode` or `None`
        """
        if user is None:
            return None
        if isinstance(user, User):
            return user.object_id
        if isinstance(user, string_types):
            return text_type(user)
        raise TypeError("Got unsupported type {0!r} for user.".format(user.__class__.__name__))

    def authenticate_instance(self, enterprise=None):
        """
        Get an access token for a Box Developer Edition enterprise.

        :param enterprise:
            The ID of the Box Developer Edition enterprise.

            Optional if the value was already given to `__init__`,
            otherwise required.
        :type enterprise:   `unicode` or `None`
        :raises:
            :exc:`ValueError` if `None` was passed for the enterprise ID here
            and in `__init__`, or if the non-`None` value passed here does not
            match the non-`None` value passed to `__init__`.
        :return:
            The access token for the enterprise which can provision/deprovision app users.
        :rtype:
            `unicode`
        """
        enterprises = [enterprise, self._enterprise_id]
        if not any(enterprises):
            raise ValueError("authenticate_instance: Requires the enterprise ID, but it was not provided.")
        if all(enterprises) and (enterprise != self._enterprise_id):
            raise ValueError(
                "authenticate_instance: Given enterprise ID {given_enterprise!r}, but {auth} already has ID {existing_enterprise!r}"
                .format(auth=self, given_enterprise=enterprise, existing_enterprise=self._enterprise_id)
            )
        if not self._enterprise_id:
            self._enterprise_id = enterprise
        self._user_id = None
        return self._auth_with_jwt(self._enterprise_id, 'enterprise')

    def _refresh(self, access_token):
        """
        Base class override.

        Instead of refreshing an access token using a refresh token, we just issue a new JWT request.
        """
        # pylint:disable=unused-argument
        if self._user_id is None:
            new_access_token = self.authenticate_instance()
        else:
            new_access_token = self.authenticate_user()
        return new_access_token, None

    @classmethod
    def _normalize_rsa_private_key(cls, file_sys_path, data, passphrase=None):
        if len(list(filter(None, [file_sys_path, data]))) != 1:
            raise TypeError("must pass exactly one of either rsa_private_key_file_sys_path or rsa_private_key_data")
        if file_sys_path:
            with open(file_sys_path, 'rb') as key_file:
                data = key_file.read()
        if hasattr(data, 'read') and callable(data.read):
            data = data.read()
        if isinstance(data, text_type):
            try:
                data = data.encode('ascii')
            except UnicodeError:
                raise_from(
                    TypeError("rsa_private_key_data must contain binary data (bytes/str), not a text/unicode string"),
                    None,
                )
        if isinstance(data, binary_type):
            passphrase = cls._normalize_rsa_private_key_passphrase(passphrase)
            return serialization.load_pem_private_key(
                data,
                password=passphrase,
                backend=default_backend(),
            )
        if isinstance(data, RSAPrivateKey):
            return data
        raise TypeError(
            'rsa_private_key_data must be binary data (bytes/str), '
            'a file-like object with a read() method, '
            'or an instance of RSAPrivateKey, '
            'but got {0!r}'
            .format(data.__class__.__name__)
        )

    @staticmethod
    def _normalize_rsa_private_key_passphrase(passphrase):
        if isinstance(passphrase, text_type):
            try:
                return passphrase.encode('ascii')
            except UnicodeError:
                raise_from(
                    TypeError("rsa_private_key_passphrase must contain binary data (bytes/str), not a text/unicode string"),
                    None,
                )
        if not isinstance(passphrase, (binary_type, NoneType)):
            raise TypeError(
                "rsa_private_key_passphrase must contain binary data (bytes/str), got {0!r}"
                .format(passphrase.__class__.__name__)
            )
        return passphrase

    @classmethod
    def from_settings_dictionary(cls, settings_dictionary, **kwargs):
        """
        Create an auth instance as defined by the given settings dictionary.

        The dictionary should have the structure of the JSON file downloaded from the Box Developer Console.

        :param settings_dictionary:       Dictionary containing settings for configuring app auth.
        :type settings_dictionary:        `dict`
        :return:                        Auth instance configured as specified by the config dictionary.
        :rtype:                         :class:`JWTAuth`
        """
        if 'boxAppSettings' not in settings_dictionary:
            raise ValueError('boxAppSettings not present in configuration')
        return cls(
            client_id=settings_dictionary['boxAppSettings']['clientID'],
            client_secret=settings_dictionary['boxAppSettings']['clientSecret'],
            enterprise_id=settings_dictionary.get('enterpriseID', None),
            jwt_key_id=settings_dictionary['boxAppSettings']['appAuth'].get('publicKeyID', None),
            rsa_private_key_data=settings_dictionary['boxAppSettings']['appAuth'].get('privateKey', None),
            rsa_private_key_passphrase=settings_dictionary['boxAppSettings']['appAuth'].get('passphrase', None),
            **kwargs
        )

    @classmethod
    def from_settings_file(cls, settings_file_sys_path, **kwargs):
        """
        Create an auth instance as defined by a JSON file downloaded from the Box Developer Console.
        See https://developer.box.com/v2.0/docs/authentication-with-jwt for more information.

        :param settings_file_sys_path:    Path to the JSON file containing the configuration.
        :type settings_file_sys_path:     `unicode`
        :return:                        Auth instance configured as specified by the JSON file.
        :rtype:                         :class:`JWTAuth`
        """
        with open(settings_file_sys_path) as config_file:
            config_dictionary = json.load(config_file)
            return cls.from_settings_dictionary(config_dictionary, **kwargs)



# coding: utf-8

from __future__ import unicode_literals

from boxsdk.object.event import Event


def test_init_event():
    event = Event(
        {
            "type": "event",
            "event_id": "f82c3ba03e41f7e8a7608363cc6c0390183c3f83",
            "source":
            {
                "type": "folder",
                "id": "11446498",
            },
        })
    assert event['type'] == 'event'
    assert event['event_id'] == 'f82c3ba03e41f7e8a7608363cc6c0390183c3f83'



# coding: utf-8

from __future__ import unicode_literals, absolute_import


from mock import mock_open


def streamable_mock_open(mock=None, read_data=b''):
    mock = mock_open(mock, read_data)
    handle = mock.return_value
    handle.position = 0

    def tell():
        return handle.position

    def read(size=-1):
        if size == -1:
            handle.position = len(read_data)
            return read_data

        data = read_data[handle.position:handle.position + size]
        handle.position += size
        return data

    # pylint:disable=no-member
    handle.tell.side_effect = tell
    handle.len = len(read_data)
    handle.read.side_effect = read
    del handle.getvalue
    return mock



# coding: utf-8

from __future__ import unicode_literals
import codecs
import json
from test.functional.mock_box.db_model.file_model import FileModel
from test.functional.mock_box.db_model import DbModel


def serializer_factory(cls):

    class BoxObjectSerializer(json.JSONEncoder):
        # pylint:disable=method-hidden
        _blacklist = ['metadata', 'files', 'folders']
        _cls = cls

        def default(self, o):
            if hasattr(o, 'isoformat'):
                return o.isoformat()
            if isinstance(o, DbModel):
                if DbModel not in self._cls.__bases__:
                    self._cls = o.__class__
                fields = {}
                for field in (x for x in dir(o) if not x.startswith('_') and x not in self._blacklist):
                    data = o.__getattribute__(field)
                    if isinstance(o, self._cls):
                        fields[field] = data
                    else:
                        try:
                            json.dumps(data)
                            fields[field] = data
                        except (TypeError, UnicodeDecodeError):
                            fields[field] = None
                object_type = type(o).__name__.lower()[:-5]
                fields['type'] = object_type
                object_id = object_type + '_id'
                if hasattr(o, 'id') and hasattr(o, object_id):
                    fields['id'] = o.__getattribute__(object_id)
                    del fields[object_id]
                if isinstance(o, FileModel):
                    del fields['content']
                return fields
            return super(BoxObjectSerializer, self).default(o)

    return BoxObjectSerializer


def loads(string, **kwargs):
    return json.loads(string, **kwargs)


def load(file_pointer, **kwargs):
    reader = codecs.getreader('utf-8')
    return json.load(reader(file_pointer), **kwargs)


def dump(obj, file_pointer, **kwargs):
    return json.dump(obj, file_pointer, cls=serializer_factory(obj.__class__), check_circular=False, **kwargs)


def dumps(obj, **kwargs):
    return json.dumps(obj, cls=serializer_factory(obj.__class__), check_circular=False, **kwargs)



# coding: utf-8

from __future__ import unicode_literals

import time

import requests

from .network_interface import Network, NetworkResponse


class DefaultNetwork(Network):
    """Implementation of the network interface using the requests library."""

    def __init__(self):
        super(DefaultNetwork, self).__init__()
        self._session = requests.Session()

    def request(self, method, url, access_token, **kwargs):
        """Base class override.
        Make a network request using a requests.Session.
        """
        return self.network_response_constructor(
            request_response=self._session.request(method, url, **kwargs),
            access_token_used=access_token,
        )

    def retry_after(self, delay, request_method, *args, **kwargs):
        """Base class override.
        Retry after sleeping for delay seconds.
        """
        time.sleep(delay)
        return request_method(*args, **kwargs)

    @property
    def network_response_constructor(self):
        """Baseclass override.

        A callable that accepts `request_response` and `access_token_used`
        keyword arguments for the :class:`DefaultNetworkResponse` constructor,
        and returns an instance of :class:`DefaultNetworkResponse`.
        """
        return DefaultNetworkResponse


class DefaultNetworkResponse(NetworkResponse):
    """Implementation of the network interface using the requests library."""

    def __init__(self, request_response, access_token_used):
        self._request_response = request_response
        self._access_token_used = access_token_used

    def json(self):
        """Base class override."""
        return self._request_response.json()

    @property
    def content(self):
        """Base class override."""
        return self._request_response.content

    @property
    def status_code(self):
        """Base class override."""
        return self._request_response.status_code

    @property
    def ok(self):
        """Base class override."""
        # pylint:disable=invalid-name
        return self._request_response.ok

    @property
    def headers(self):
        """Base class override."""
        return self._request_response.headers

    @property
    def response_as_stream(self):
        """Base class override."""
        return self._request_response.raw

    @property
    def access_token_used(self):
        """Base class override."""
        return self._access_token_used

    @property
    def request_response(self):
        """
        The response returned from the Requests library.

        :rtype: `Response`
        """
        return self._request_response



# coding: utf-8

from __future__ import unicode_literals

from hashlib import sha1

from bottle import response, request
from sqlalchemy.orm import make_transient

from test.functional.mock_box.behavior.item_behavior import ItemBehavior
from test.functional.mock_box.db_model.event_model import EventModel
from test.functional.mock_box.db_model.file_model import FileModel
from test.functional.mock_box.util.db_utils import get_file_by_id, get_folder_by_id, get_user_from_header
from test.functional.mock_box.util.http_utils import abort
from test.functional.mock_box.util import json_utils as json


class FileBehavior(ItemBehavior):
    @staticmethod
    def _check_file_lock(file_object, is_download=False):
        if len(file_object.locks):
            lock_object = file_object.locks[0]
            if not is_download or lock_object.is_download_prevented:
                abort(403, 'File is locked.')

    def update_file(self, file_id):
        """
        https://developers.box.com/docs/#files-upload-a-new-version-of-a-file
        """
        file_object = get_file_by_id(self._db_session, file_id)
        self._check_file_lock(file_object)
        content = request.files.file.file.read()
        file_hash = sha1()
        file_hash.update(content)
        file_object.content = content
        file_object.sha1 = file_hash.hexdigest()
        file_object.size = len(content)
        self._db_session.add(EventModel(event_type='ITEM_UPLOAD', source_id=file_object.file_id, source_type='file'))
        self._db_session.commit()
        return json.dumps({'entries': [file_object]})

    def upload_file(self):
        """
        https://developers.box.com/docs/#files-upload-a-file
        """
        attributes = request.forms.get('attributes') or request.forms.get('metadata')
        if not attributes:
            abort(400, 'Missing parameter: attributes')
        attributes = json.loads(attributes)
        parent = attributes.get('parent')
        if parent is None or 'id' not in parent:
            abort(400, 'Missing parameter: parent(id)')
        parent_id = parent['id']
        folder = get_folder_by_id(self._db_session, parent_id)
        content = request.files.file.file.read()
        file_hash = sha1()
        file_hash.update(content)
        owner = get_user_from_header(self._db_session)
        file_object = FileModel(
            content=content,
            name=attributes.get('name', request.files.file.name),
            parent_id=folder.id,
            sha1=file_hash.hexdigest(),
            size=len(content),
            owned_by=owner,
            created_by=owner,
        )
        self._db_session.add(file_object)
        self._db_session.commit()
        self._db_session.add(EventModel(event_type='ITEM_UPLOAD', source_id=file_object.file_id, source_type='file'))
        self._db_session.commit()
        return json.dumps({'entries': [file_object]})

    def get_file_info(self, file_id):
        """
        https://developers.box.com/docs/#files-get
        """
        file_object = get_file_by_id(self._db_session, file_id)
        return json.dumps(file_object)

    def update_file_info(self, file_id):
        file_object = get_file_by_id(self._db_session, file_id)
        self._check_file_lock(file_object)
        params = json.load(request.body)
        for key, value in params.items():
            if not hasattr(FileModel, key):
                abort(400, 'File has no attribute {0}.'.format(key))
            if key == 'parent':
                # Move
                parent_id = value['id']
                parent_folder = get_folder_by_id(self._db_session, parent_id)
                file_object.parent_id = parent_folder.id
                self._db_session.add(
                    EventModel(event_type='ITEM_MOVE', source_id=file_object.file_id, source_type='file'),
                )
            else:
                setattr(file_object, key, value)
                if key == 'name':
                    self._db_session.add(
                        EventModel(event_type='ITEM_RENAME', source_id=file_object.file_id, source_type='file'),
                    )
                elif key == 'sync_state':
                    event_type = 'ITEM_SYNC' if value == 'synced' else 'ITEM_UNSYNC'
                    self._db_session.add(
                        EventModel(event_type=event_type, source_id=file_object.file_id, source_type='file'),
                    )
        self._db_session.commit()
        return json.dumps(file_object)

    def copy_file(self, file_id):
        file_object = get_file_by_id(self._db_session, file_id)
        self._check_file_lock(file_object)
        parent_folder = self._get_parent()
        self._db_session.expunge(file_object)
        make_transient(file_object)
        file_object.id = None
        file_object.file_id = None
        file_object.parent_id = parent_folder.id
        self._db_session.add(file_object)
        self._db_session.commit()
        self._db_session.add(EventModel(event_type='ITEM_COPY', source_id=file_object.file_id, source_type='file'))
        self._db_session.commit()
        return json.dumps(file_object)

    def download_file(self, file_id):
        """
        https://developers.box.com/docs/#files-download-a-file
        """
        file_object = get_file_by_id(self._db_session, file_id)
        self._check_file_lock(file_object, is_download=True)
        return file_object.content

    def delete_file(self, file_id):
        file_object = get_file_by_id(self._db_session, file_id)
        self._check_file_lock(file_object)
        self._db_session.delete(file_object)
        self._db_session.add(EventModel(event_type='ITEM_TRASH', source_id=file_object.file_id, source_type='file'))
        self._db_session.commit()
        response.status = 204



# coding: utf-8

from __future__ import absolute_import, unicode_literals

from functools import partial
from pprint import pformat
import sys

from six import text_type

from boxsdk.network.default_network import DefaultNetwork, DefaultNetworkResponse
from boxsdk.util.log import setup_logging


class LoggingNetwork(DefaultNetwork):
    """
    SDK Network subclass that logs requests and responses.
    """
    LOGGER_NAME = 'boxsdk.network'
    REQUEST_FORMAT = '\x1b[36m%(method)s %(url)s %(request_kwargs)s\x1b[0m'
    EXCEPTION_FORMAT = '\x1b[31mRequest "%(method)s %(url)s" failed with %(exc_type_name)s exception: %(exc_value)r\x1b[0m'
    _COMMON_RESPONSE_FORMAT = '"%(method)s %(url)s" %(status_code)s %(content_length)s\n%(headers)s\n%(content)s\n'
    SUCCESSFUL_RESPONSE_FORMAT = '\x1b[32m{0}\x1b[0m'.format(_COMMON_RESPONSE_FORMAT)
    ERROR_RESPONSE_FORMAT = '\x1b[31m{0}\x1b[0m'.format(_COMMON_RESPONSE_FORMAT)

    def __init__(self, logger=None):
        """
        :param logger:
            The logger to use. If you instantiate this class more than once, you should use the same logger
            to avoid duplicate log entries.
        :type logger:
            :class:`Logger`
        """
        super(LoggingNetwork, self).__init__()
        self._logger = logger or setup_logging(name=self.LOGGER_NAME)

    @property
    def logger(self):
        return self._logger

    def _log_request(self, method, url, **kwargs):
        """
        Logs information about the Box API request.

        :param method:
            The HTTP verb that should be used to make the request.
        :type method:
            `unicode`
        :param url:
            The URL for the request.
        :type url:
            `unicode`
        :param access_token:
            The OAuth2 access token used to authorize the request.
        :type access_token:
            `unicode`
        """
        self._logger.info(self.REQUEST_FORMAT, {'method': method, 'url': url, 'request_kwargs': pformat(kwargs)})

    def _log_exception(self, method, url, exc_info):
        """Log information at WARNING level about the exception that was raised when trying to make the request.

        :param method:  The HTTP verb that was used to make the request.
        :type method:   `unicode`
        :param url:   The URL for the request.
        :type url:  `unicode`
        :param exc_info:  The exception info returned from `sys.exc_info()`.
        """
        exc_type, exc_value, _ = exc_info
        self._logger.warning(
            self.EXCEPTION_FORMAT,
            {'method': method, 'url': url, 'exc_type_name': exc_type.__name__, 'exc_value': exc_value},
        )

    def request(self, method, url, access_token, **kwargs):
        """
        Base class override. Logs information about an API request and response in addition to making the request.

        Also logs exceptions before re-raising them.

        The logging of the response is deferred to
        :class:`LoggingNetworkResponse`. See that class's docstring for more
        info.
        """
        self._log_request(method, url, **kwargs)
        try:
            return super(LoggingNetwork, self).request(method, url, access_token, **kwargs)
        except Exception:
            self._log_exception(method, url, sys.exc_info())
            raise

    @property
    def network_response_constructor(self):
        """Baseclass override.

        A callable that passes additional required keyword arguments to the
        :class:`LoggingNetworkResponse` constructor, and returns an instance of
        :class:`LoggingNetworkResponse`.
        """
        return partial(
            LoggingNetworkResponse,
            logger=self._logger,
            successful_response_format=self.SUCCESSFUL_RESPONSE_FORMAT,
            error_response_format=self.ERROR_RESPONSE_FORMAT,
        )


class LoggingNetworkResponse(DefaultNetworkResponse):
    """Response subclass that defers LoggingNetwork response logging until it is safe to do so.

    :class:`DefaultNetwork` is based off the `requests` library.
    :class:`requests.Response` has a few mutually-exclusive ways to read the
    content of the response:

        - With the `Response.raw` attribute, an `io.IOBase` instance returned
          from the `urllib3` library, that can be read once in chunks from
          beginning to end.
        - With `Response.iter_content()` and other iter_* generators, which
          also can only be read once and advance the `Response.raw` IO stream.
        - With the `Response.content` property (and other attributes such as
          `Response.text` and `Response.json()`), which reads and caches the
          remaining response content in memory. Can be accessed multiple times,
          but cannot be safely accessed if any of the previous mechanisms have
          been used at all. And if this property has already been accessed,
          then the other mechanisms will have been exhausted, and attempting to
          read from them will make it appear like the response content is
          empty.

    Any of these mechanisms may be used to read any response, regardless of
    whether `stream=True` or `stream=False` on the request.

    If the caller uses `Response.content`, then it is safe for
    :class:`LoggingNetwork` to also access it. But if the caller uses any of
    the streaming mechanisms, then it is not safe for :class:`LoggingNetwork`
    to ever read any of the content. Thus, the options available are:

        - Never log the content of a response.
        - Make logging part of the :class:`Network` interface, and add an
          optional keyword argument that callers can use to specify when it is
          unsafe to log the content of a response.
        - Defer logging until it is possible to auto-detect which mechanism is
          being used.

    This class is an implementation of the latter option. Instead of response
    logging taking place in `LoggingNetwork.request()`, it takes place in this
    `DefaultNetworkResponse` subclass, as soon as the caller starts reading the
    content. If `content` or `json()` are accessed, then the response will be
    logged with its content. Whereas if `response_as_stream` or
    `request_response` are accessed, then the response will be logged with a
    placeholder for the actual content.

    In theory, this could make the logs less useful, by adding a delay between
    when the network response was actually received, and when it is logged. Or
    the response may never be logged, if the content is never accessed. In
    practice, this is unlikely to happen, because nearly all SDK methods
    immediately read the content.
    """

    STREAM_CONTENT_NOT_LOGGED = '<File download contents unavailable for logging>'

    def __init__(self, logger, successful_response_format, error_response_format, **kwargs):
        """Extends baseclass method.

        :param logger:  The logger to use.
        :type logger:   :class:`Logger`
        :param successful_response_format:
            The logger %-style format string to use for logging ok responses.

            May use the following format placeholders:

                - %(method)s : The HTTP request method ('GET', 'POST', etc.).
                - %(url)s : The url of the request.
                - %(status_code)s : The HTTP status code of the response.
                - %(content_length)s : The Content-Length of the response body.
                - %(headers)s : The HTTP headers (as a pretty-printed dict).
                - %(content)s : The response body.

        :type successful_response_format:   `unicode`
        :param error_response_format:
            The logger %-style format string to use for logging ok responses.

            May use the same format placeholders as
            `successful_response_format`.
        :type error_response_format:  `unicode`
        """
        super(LoggingNetworkResponse, self).__init__(**kwargs)
        self._logger = logger
        self._successful_response_format = successful_response_format
        self._error_response_format = error_response_format
        self._did_log = False

    def log(self, can_safely_log_content=False):
        """Logs information about the Box API response.

        Will only execute once. Subsequent calls will be no-ops. This is
        partially because we only want to log responses once, and partially
        because this is necessary to prevent this method from infinite
        recursing with its use of the `content` property.

        :param can_safely_log_content:
            (optional) `True` if the caller is accessing the `content`
            property, `False` otherwise.

            As stated in the class docstring, it is unsafe for this logging
            method to access `content` unless the caller is also accessing it.

            Defaults to `False`.
        :type can_safely_log_content:   `bool`
        """
        if self._did_log:
            return
        self._did_log = True
        content_length = self.headers.get('Content-Length', None)
        content = self.STREAM_CONTENT_NOT_LOGGED
        if can_safely_log_content:
            if content_length is None:
                content_length = text_type(len(self.content))

            # If possible, get the content as a JSON `dict`, that way
            # `pformat(content)` will return pretty-printed JSON.
            try:
                content = self.json()
            except ValueError:
                content = self.content
            content = pformat(content)
        if content_length is None:
            content_length = '?'
        if self.ok:
            logger_method, response_format = self._logger.info, self._successful_response_format
        else:
            logger_method, response_format = self._logger.warning, self._error_response_format
        logger_method(
            response_format,
            {
                'method': self.request_response.request.method,
                'url': self.request_response.request.url,
                'status_code': self.status_code,
                'content_length': content_length,
                'headers': pformat(self.headers),
                'content': content,
            },
        )

    def json(self):
        """Extends baseclass method."""
        try:
            return super(LoggingNetworkResponse, self).json()
        finally:
            self.log(can_safely_log_content=True)

    @property
    def content(self):
        """Extends baseclass method."""
        content = super(LoggingNetworkResponse, self).content
        self.log(can_safely_log_content=True)
        return content

    @property
    def response_as_stream(self):
        """Extends baseclass method."""
        stream = super(LoggingNetworkResponse, self).response_as_stream
        self.log(can_safely_log_content=False)
        return stream

    @property
    def request_response(self):
        """Extends baseclass method."""
        response = super(LoggingNetworkResponse, self).request_response
        self.log(can_safely_log_content=False)
        return response



# coding: utf-8

from __future__ import unicode_literals, absolute_import
import pytest
from boxsdk.object.folder import FolderSyncState


@pytest.fixture()
def extra_network_parameters():
    return {'timeout': 1}


def test_folder_clone_during_create_subfolder(created_subfolder, extra_network_parameters):
    # pylint:disable=redefined-outer-name
    # pylint:disable=protected-access
    original_folder = created_subfolder
    original_folder_id = original_folder._object_id
    original_folder_response_object = original_folder._response_object
    original_session = original_folder._session
    returned_folder = original_folder.create_subfolder('subfolder', extra_network_parameters=extra_network_parameters)
    returned_session = returned_folder._session

    assert original_session is not returned_session
    assert original_session._default_network_request_kwargs == {}
    assert returned_session._default_network_request_kwargs == {'timeout': 1}
    assert original_folder._object_id == original_folder_id
    assert original_folder._response_object == original_folder_response_object


def test_folder_clone_during_update_sync_state(created_subfolder, extra_network_parameters):
    # pylint:disable=redefined-outer-name
    # pylint:disable=protected-access
    original_folder = created_subfolder
    original_folder_id = original_folder._object_id
    original_folder_response_object = original_folder._response_object
    original_session = original_folder._session
    returned_folder = original_folder.update_sync_state(FolderSyncState.IS_SYNCED, extra_network_parameters=extra_network_parameters)
    returned_session = returned_folder._session

    assert original_session is not returned_session
    assert original_session._default_network_request_kwargs == {}
    assert returned_session._default_network_request_kwargs == {'timeout': 1}
    assert original_folder._object_id == original_folder_id
    assert original_folder._response_object == original_folder_response_object



# coding: utf-8

from __future__ import unicode_literals, absolute_import

from functools import partial
import json
from logging import Logger
from operator import attrgetter
from pprint import pformat

import mock
from mock import Mock, patch
import pytest
from six import text_type
from six.moves import map

from boxsdk.network import default_network, logging_network
from boxsdk.network.logging_network import LoggingNetwork, LoggingNetworkResponse


class ExceptionSubclass(Exception):
    pass


@pytest.fixture
def logger():
    return Mock(Logger)


@pytest.fixture
def network(logger):
    return LoggingNetwork(logger)


@pytest.fixture(scope='module')
def logger_method_names():
    return ['critical', 'debug', 'error', 'exception', 'fatal', 'info', 'log', 'warn', 'warning']


@pytest.fixture
def logger_methods(logger, logger_method_names):
    return list(map(partial(getattr, logger), logger_method_names))


@pytest.fixture
def assert_logger_called_once_with(logger, logger_method_names):

    def _assert_logger_called_once_with(method_name, *args, **kwargs):
        getattr(logger, method_name).assert_called_once_with(*args, **kwargs)
        method_names = set(logger_method_names)
        method_names.discard(method_name)
        for name in method_names:
            getattr(logger, name).assert_not_called()

    return _assert_logger_called_once_with


@pytest.fixture
def assert_logger_not_called(logger_methods):

    def _assert_logger_not_called():
        for method in logger_methods:
            method.assert_not_called()

    return _assert_logger_not_called


@pytest.fixture
def logger_call_count(logger_methods):

    def _logger_call_count():
        return sum(map(attrgetter('call_count'), logger_methods))

    return _logger_call_count


# BEGIN Tests for LoggingNetwork.


def test_logging_network_calls_setup_logging_if_logger_is_none():
    with patch.object(logging_network, 'setup_logging') as setup_logging:
        network = LoggingNetwork()
        setup_logging.assert_called_once_with(name=LoggingNetwork.LOGGER_NAME)
        assert network.logger is setup_logging.return_value


def test_logging_network_can_be_initialized_if_logger_is_none(logger):
    with patch('logging.getLogger') as get_logger:
        get_logger.return_value = logger
        network = LoggingNetwork()
        assert network.logger is get_logger.return_value
        get_logger.assert_called_once_with(LoggingNetwork.LOGGER_NAME)


def test_logging_network_does_not_call_setup_logging_if_logger_is_not_none(logger):
    with patch.object(logging_network, 'setup_logging') as setup_logging:
        network = LoggingNetwork(logger)
        setup_logging.assert_not_called()
        assert network.logger is logger


def test_logging_network_logs_requests(make_network_request, http_verb, test_url, network, assert_logger_called_once_with):
    kwargs = dict(custom_kwarg='foo')
    make_network_request(network, **kwargs)
    assert_logger_called_once_with(
        'info',
        network.REQUEST_FORMAT,
        {'method': http_verb, 'url': test_url, 'request_kwargs': pformat(kwargs)},
    )


def test_logging_network_logs_request_exception(make_network_request, http_verb, test_url, logger, network):
    with patch.object(default_network.DefaultNetwork, 'request') as super_request:
        super_request.side_effect = expected_exception = ExceptionSubclass('exception raised from request()')
        with pytest.raises(ExceptionSubclass) as pytest_exc_info:
            make_network_request(network)
        assert pytest_exc_info.value is expected_exception
    logger.warning.assert_called_once_with(
        network.EXCEPTION_FORMAT,
        {'method': http_verb, 'url': test_url, 'exc_type_name': ExceptionSubclass.__name__, 'exc_value': expected_exception},
    )


def test_logging_network_request_returns_logging_network_response(make_network_request, request_response, network):
    response = make_network_request(network)
    assert response.request_response is request_response
    assert isinstance(response, LoggingNetworkResponse)


def test_logging_network_response_constructor(make_network_request, access_token, logger, network):
    mock_response = Mock(LoggingNetworkResponse, access_token_used=access_token)
    with patch.object(logging_network, 'LoggingNetworkResponse') as mock_logging_network_response_class:
        mock_logging_network_response_class.return_value = mock_response
        assert make_network_request(network) is mock_response
        assert mock_logging_network_response_class.call_count == 1
        kwargs = mock_logging_network_response_class.call_args[1]
        assert ('logger' in kwargs) and (kwargs['logger'] is logger)
        assert ('successful_response_format' in kwargs) and (kwargs['successful_response_format'] is network.SUCCESSFUL_RESPONSE_FORMAT)
        assert ('error_response_format' in kwargs) and (kwargs['error_response_format'] is network.ERROR_RESPONSE_FORMAT)


# END Tests for LoggingNetwork.


# BEGIN Tests for LoggingNetworkResponse.


@pytest.fixture
def construct_logging_network_response(logger, access_token):

    def _construct_logging_network_response(request_response):
        return LoggingNetworkResponse(
            logger=logger,
            successful_response_format=LoggingNetwork.SUCCESSFUL_RESPONSE_FORMAT,
            error_response_format=LoggingNetwork.ERROR_RESPONSE_FORMAT,
            request_response=request_response,
            access_token_used=access_token,
        )

    return _construct_logging_network_response


@pytest.fixture
def logging_network_response(construct_logging_network_response, request_response):
    return construct_logging_network_response(request_response)


@pytest.fixture(params=['content', 'json'])
def get_content_from_response(request):

    def _get_content_from_response(response):
        content = getattr(response, request.param)
        if request.param in ['json']:
            try:
                content = content()
            except ValueError:
                pass

    return _get_content_from_response


@pytest.fixture(params=['response_as_stream', 'request_response', 'log'])
def do_not_get_content_from_response(request):

    def _do_not_get_content_from_response(response):
        attribute = getattr(response, request.param)
        if request.param in ['log']:
            attribute()

    return _do_not_get_content_from_response


def test_logging_network_response_does_not_log_anything_immediately(logging_network_response, logger_call_count):
    # pylint:disable=unused-argument
    # Need to load the `logging_network_response` fixture for this test to be
    # meaningful.
    assert logger_call_count() == 0


def test_logging_network_response_only_logs_once(logging_network_response, logger_call_count):
    logging_network_response.log()
    assert logger_call_count() == 1
    logging_network_response.log()
    assert logger_call_count() == 1


@pytest.mark.parametrize('content_length_header', [False, True])
def test_logging_network_logs_successful_responses(
        construct_logging_network_response, generic_successful_request_response, assert_logger_called_once_with,
        get_content_from_response, http_verb, test_url, content_length_header,
):
    expected_content_length = text_type(len(generic_successful_request_response.content))
    if content_length_header:
        generic_successful_request_response.headers['Content-Length'] = expected_content_length
    else:
        generic_successful_request_response.headers.pop('Content-Length', None)
    generic_successful_request_response.request.method = http_verb
    generic_successful_request_response.request.url = test_url
    logging_network_response = construct_logging_network_response(generic_successful_request_response)
    get_content_from_response(logging_network_response)
    assert_logger_called_once_with(
        'info',
        LoggingNetwork.SUCCESSFUL_RESPONSE_FORMAT,
        {
            'method': http_verb,
            'url': test_url,
            'status_code': generic_successful_request_response.status_code,
            'content_length': expected_content_length,
            'headers': pformat(generic_successful_request_response.headers),
            'content': pformat(generic_successful_request_response.json()),
        }
    )


@pytest.mark.parametrize('content_length_header', [False, True])
def test_logging_network_logs_successful_responses_with_stream_placeholder(
        construct_logging_network_response, generic_successful_request_response, assert_logger_called_once_with,
        do_not_get_content_from_response, logger, content_length_header,
):
    if content_length_header:
        expected_content_length = text_type(len(generic_successful_request_response.content))
        generic_successful_request_response.headers['Content-Length'] = expected_content_length
    else:
        generic_successful_request_response.headers.pop('Content-Length', None)
        expected_content_length = '?'
    logging_network_response = construct_logging_network_response(generic_successful_request_response)
    do_not_get_content_from_response(logging_network_response)
    assert_logger_called_once_with('info', LoggingNetwork.SUCCESSFUL_RESPONSE_FORMAT, mock.ANY)
    assert logger.info.call_args[0][1]['content'] == logging_network_response.STREAM_CONTENT_NOT_LOGGED
    assert logger.info.call_args[0][1]['content_length'] == expected_content_length


def test_logging_network_logs_successful_responses_with_non_json_content(
        construct_logging_network_response, generic_successful_request_response, assert_logger_called_once_with,
        logger, get_content_from_response,
):
    generic_successful_request_response.content = content = (b''.join(chr(i).encode('utf-8') for i in range(128)) * 4)
    generic_successful_request_response.json.side_effect = lambda: json.loads(content.decode('utf-8'))
    logging_network_response = construct_logging_network_response(generic_successful_request_response)
    get_content_from_response(logging_network_response)
    assert_logger_called_once_with('info', LoggingNetwork.SUCCESSFUL_RESPONSE_FORMAT, mock.ANY)
    assert logger.info.call_args[0][1]['content'] == pformat(content)


def test_logging_network_logs_non_successful_responses(
        construct_logging_network_response, server_error_request_response, assert_logger_called_once_with,
        get_content_from_response, http_verb, test_url,
):
    server_error_request_response.request.method = http_verb
    server_error_request_response.request.url = test_url
    logging_network_response = construct_logging_network_response(server_error_request_response)
    get_content_from_response(logging_network_response)
    assert_logger_called_once_with(
        'warning',
        LoggingNetwork.ERROR_RESPONSE_FORMAT,
        {
            'method': http_verb,
            'url': test_url,
            'status_code': server_error_request_response.status_code,
            'content_length': text_type(len(server_error_request_response.content)),
            'headers': pformat(server_error_request_response.headers),
            'content': pformat(server_error_request_response.json()),
        }
    )


# END Tests for LoggingNetworkResponse.



# coding: utf-8

from __future__ import unicode_literals
import pytest
from boxsdk.client import Client
from boxsdk.exception import BoxAPIException


def test_create_folder_then_update_info(created_subfolder):
    # pylint:disable=redefined-outer-name
    _test_create_then_update_info(created_subfolder)


def test_create_file_then_update_info(uploaded_file):
    # pylint:disable=redefined-outer-name
    _test_create_then_update_info(uploaded_file)


def _test_create_then_update_info(item):
    updated_name = 'updated_{0}'.format(item.name)
    updated_item = item.update_info({'name': updated_name})
    assert updated_item.name == updated_name
    assert item.get().name == updated_name


def test_create_folder_then_rename(created_subfolder):
    # pylint:disable=redefined-outer-name
    _test_create_then_rename(created_subfolder)


def test_create_file_then_rename(uploaded_file):
    # pylint:disable=redefined-outer-name
    _test_create_then_rename(uploaded_file)


def _test_create_then_rename(item):
    updated_name = 'updated_{0}'.format(item.name)
    updated_item = item.rename(updated_name)
    assert updated_item.name == updated_name
    assert item.get().name == updated_name


def test_create_folder_then_move(box_client, created_subfolder):
    # pylint:disable=redefined-outer-name
    _test_create_then_move(box_client, created_subfolder)


def test_create_file_then_move(box_client, uploaded_file):
    # pylint:disable=redefined-outer-name
    _test_create_then_move(box_client, uploaded_file)


def _test_create_then_move(box_client, item):
    item_name = item.name
    move_target = box_client.folder('0').create_subfolder('move target')
    item.move(move_target)
    item = item.get()
    assert item.name == item_name
    assert item.parent['id'] == move_target.object_id
    assert len(box_client.folder('0').get_items(10)) == 1
    assert len(move_target.get_items(10)) == 1


def test_create_folder_then_copy(box_client, created_subfolder):
    # pylint:disable=redefined-outer-name
    _test_create_then_copy(box_client, created_subfolder)


def test_create_file_then_copy(box_client, uploaded_file):
    # pylint:disable=redefined-outer-name
    _test_create_then_copy(box_client, uploaded_file)


def _test_create_then_copy(box_client, item):
    # pylint:disable=redefined-outer-name
    copy_target = box_client.folder('0').create_subfolder('copy target')
    copied_item = item.copy(copy_target)
    item = item.get()
    copied_item = copied_item.get()
    assert item.id != copied_item.id
    assert item.name == copied_item.name
    assert copied_item.parent['id'] == copy_target.object_id
    assert len(box_client.folder('0').get_items(10)) == 2
    assert len(copy_target.get_items(10)) == 1


@pytest.mark.parametrize('constructor', [Client.file, Client.folder])
def test_get_item_info_for_missing_item(box_client, constructor):
    with pytest.raises(BoxAPIException) as exc_info:
        constructor(box_client, '1').get()
    assert exc_info.value.status == 404


@pytest.mark.parametrize('sync_state', ['synced', 'not_synced'])
def test_set_sync_state(created_subfolder, sync_state):
    # pylint:disable=redefined-outer-name
    folder_with_info = created_subfolder.get(fields='sync_state')
    created_subfolder.update_sync_state(sync_state)
    updated_folder_with_info = created_subfolder.get(fields='sync_state')
    assert folder_with_info.sync_state == 'not_synced'
    assert updated_folder_with_info.sync_state == sync_state



# coding: utf-8

from __future__ import unicode_literals, absolute_import

from ..auth import DeveloperTokenAuth
from .client import Client


class DeveloperTokenClient(Client):
    """
    Box client subclass which authorizes with a developer token.
    """
    def __init__(self, oauth=None, network_layer=None, session=None):
        super(DeveloperTokenClient, self).__init__(
            oauth=oauth or DeveloperTokenAuth(),
            network_layer=network_layer,
            session=session,
        )



# coding: utf-8

from __future__ import unicode_literals

from uuid import uuid4

from redis import StrictRedis
from redis.lock import Lock

from .oauth2 import OAuth2


class RedisManagedOAuth2Mixin(OAuth2):
    """
    Box SDK OAuth2 subclass.
    Allows for storing auth tokens in redis.

    :param unique_id:
        An identifier for this auth object. Auth instances which wish to share tokens must use the same ID.
    :type unique_id:
        `unicode`
    :param redis_server:
        An instance of a Redis server, configured to talk to Redis.
    :type redis_server:
        :class:`Redis`
    """
    def __init__(self, unique_id=uuid4(), redis_server=None, *args, **kwargs):
        self._unique_id = unique_id
        self._redis_server = redis_server or StrictRedis()
        refresh_lock = Lock(redis=self._redis_server, name='{0}_lock'.format(self._unique_id))
        super(RedisManagedOAuth2Mixin, self).__init__(*args, refresh_lock=refresh_lock, **kwargs)
        if self._access_token is None:
            self._get_and_update_current_tokens()

    @property
    def unique_id(self):
        """
        Get the unique ID used by this auth instance. Other instances can share tokens with this instance
        if they share the ID with this instance.
        """
        return self._unique_id

    def _get_tokens(self):
        """
        Base class override.
        Gets the latest tokens from redis before returning them.
        """
        return self._redis_server.hvals(self._unique_id) or (None, None)

    def _store_tokens(self, access_token, refresh_token):
        """
        Base class override.
        Saves the refreshed tokens in redis.
        """
        super(RedisManagedOAuth2Mixin, self)._store_tokens(access_token, refresh_token)
        self._redis_server.hmset(self._unique_id, {'access': access_token, 'refresh': refresh_token})


class RedisManagedOAuth2(RedisManagedOAuth2Mixin):
    """
    OAuth2 subclass which uses Redis to manage tokens.
    """
    pass



# coding: utf-8

from __future__ import unicode_literals
from datetime import datetime, timedelta
import pytest
from boxsdk.util.compat import total_seconds, with_metaclass


@pytest.fixture(params=(
    (timedelta(seconds=7), 7),
    (datetime(2015, 7, 6, 12) - datetime(2015, 7, 6, 11), 60 * 60),
    (timedelta(minutes=1), 60),
))
def total_seconds_data(request):
    return request.param


def test_total_seconds(total_seconds_data):
    # pylint:disable=redefined-outer-name
    delta, seconds = total_seconds_data
    assert total_seconds(delta) == seconds


def test_with_metaclass():

    class Class1(object):
        pass

    class Class2(object):
        pass

    bases = (Class1, Class2)

    class Meta(type):
        @classmethod
        def __prepare__(metacls, name, this_bases, **kwds):   # pylint:disable=unused-argument
            assert this_bases == bases
            return {}

        def __new__(metacls, name, this_bases, namespace, **kwds):
            assert this_bases == bases
            return super(Meta, metacls).__new__(metacls, name, this_bases, namespace, **kwds)

    temporary_class = with_metaclass(Meta, *bases)
    assert isinstance(temporary_class, Meta)
    assert temporary_class.__bases__ == bases

    class Subclass(temporary_class):
        pass

    assert type(Subclass) is Meta   # pylint:disable=unidiomatic-typecheck
    assert Subclass.__bases__ == bases



# coding: utf-8

from __future__ import unicode_literals
from mock import patch
import six
from test.util.streamable_mock_open import streamable_mock_open


def test_upload_then_update(box_client, test_file_path, test_file_content, update_file_content, file_name):
    with patch('boxsdk.object.folder.open', streamable_mock_open(read_data=test_file_content), create=True):
        file_object = box_client.folder('0').upload(test_file_path, file_name)
    assert file_object.name == file_name
    file_object_with_info = file_object.get()
    assert file_object_with_info.id == file_object.object_id
    assert file_object_with_info.name == file_name
    file_content = file_object.content()
    expected_file_content = test_file_content.encode('utf-8') if isinstance(test_file_content, six.text_type)\
        else test_file_content
    assert file_content == expected_file_content
    folder_items = box_client.folder('0').get_items(100)
    assert len(folder_items) == 1
    assert folder_items[0].object_id == file_object.object_id
    assert folder_items[0].name == file_object.name
    with patch('boxsdk.object.file.open', streamable_mock_open(read_data=update_file_content), create=True):
        updated_file_object = file_object.update_contents(test_file_path)
    assert updated_file_object.name == file_name
    file_object_with_info = updated_file_object.get()
    assert file_object_with_info.id == updated_file_object.object_id
    assert file_object_with_info.name == file_name
    file_content = updated_file_object.content()
    expected_file_content = update_file_content.encode('utf-8') if isinstance(update_file_content, six.text_type)\
        else update_file_content
    assert file_content == expected_file_content
    folder_items = box_client.folder('0').get_items(100)
    assert len(folder_items) == 1
    assert folder_items[0].object_id == file_object.object_id
    assert folder_items[0].name == file_object.name


def test_upload_then_download(box_client, test_file_path, test_file_content, file_name):
    with patch('boxsdk.object.folder.open', streamable_mock_open(read_data=test_file_content), create=True):
        file_object = box_client.folder('0').upload(test_file_path, file_name)
    writeable_stream = six.BytesIO()
    file_object.download_to(writeable_stream)
    expected_file_content = test_file_content.encode('utf-8') if isinstance(test_file_content, six.text_type)\
        else test_file_content
    assert writeable_stream.getvalue() == expected_file_content



# coding: utf-8
# pylint:disable=no-value-for-parameter

from __future__ import absolute_import, unicode_literals

from itertools import chain
import sys

from enum import EnumMeta
from six import reraise
from six.moves import map   # pylint:disable=redefined-builtin

from .ordered_dict import OrderedDict


__all__ = list(map(str, ['ExtendableEnumMeta']))


class ExtendableEnumMeta(EnumMeta):
    """A metaclass for enum hierarchies.

    This allows you to define hierarchies such as this:

        from box.util.compat import with_metaclass

        class EnumBase(with_metaclass(ExtendableEnumMeta, Enum)): pass

        class Enum1(EnumBase):
            A = 'A'

        class Enum2(EnumBase): pass

        class Enum2_1(Enum2):
            B = 'B'

        class Enum2_2(Enum2):
            C = 'C'

    and have all members be accessible on EnumBase (as well as have all members
    of Enum2_1 and Enum2_2 be available on Enum2) as if they had been defined
    there.

    Non-leaf classes still may not have members directly defined on them, as
    with standard enums.

    Most of the usual enum magic methods are extended: __contains__, __dir__,
    __getitem__, __getattr__, __iter__, __len__, and __reversed__. Only __new__
    is not extended; instead, a new method `lookup` is provided. The
    __members__ property is also extended.
    """

    def lookup(cls, value):
        """Custom value lookup, which does recursive lookups on subclasses.

        If this is a leaf enum class with defined members, this acts the same
        as __new__().

        But if this is a base class with no defined members of its own, it
        tries doing a value lookup on all its subclasses until it finds the
        value.

        NOTE: Because of the implementation details of Enum, this must be a new
        classmethod, and can't be implemented as __new__() [1].

        [1] <https://docs.python.org/3.5/library/enum.html#finer-points>

        :param value:
            The value to look up. Can be a value, or an enum instance.
        :type value:
            `varies`
        :raises:
            :class:`ValueError` if the value isn't found anywhere.
        """
        try:
            return cls(value)
        except ValueError:
            exc_info = sys.exc_info()
            for subclass in cls.__subclasses__():
                try:
                    return subclass.lookup(value)
                except ValueError:
                    pass
            # This needs to be `reraise()`, and not just `raise`. Otherwise,
            # the inner exception from the previous line is re-raised, which
            # isn't desired.
            reraise(*exc_info)

    @property
    def __members__(cls):
        members = OrderedDict(super(ExtendableEnumMeta, cls).__members__)
        for subclass in cls.__subclasses__():
            members.update(subclass.__members__)
        return members

    def __contains__(cls, member):
        if super(ExtendableEnumMeta, cls).__contains__(member):
            return True

        def in_(subclass):
            return member in subclass

        return any(map(in_, cls.__subclasses__()))

    def __dir__(cls):
        return list(set(super(ExtendableEnumMeta, cls).__dir__()).union(*map(dir, cls.__subclasses__())))

    def __getitem__(cls, name):
        try:
            return super(ExtendableEnumMeta, cls).__getitem__(name)
        except KeyError:
            exc_info = sys.exc_info()
            for subclass in cls.__subclasses__():
                try:
                    return subclass[name]
                except KeyError:
                    pass
            # This needs to be `reraise()`, and not just `raise`. Otherwise,
            # the inner exception from the previous line is re-raised, which
            # isn't desired.
            reraise(*exc_info)

    def __getattr__(cls, name):
        try:
            return super(ExtendableEnumMeta, cls).__getattr__(name)
        except AttributeError:
            exc_info = sys.exc_info()
            try:
                # If the super() call fails, don't call getattr() on all of the
                # subclasses. Instead, use __getitem__ to do this. This is
                # because we don't want to grab arbitrary attributes from
                # subclasses, only enum members. For enum members, __getattr__
                # and __getitem__ have the same behavior. And __getitem__ has
                # the advantage of never grabbing anything other than enum
                # members.
                return cls[name]  # pylint:disable=unsubscriptable-object
            except KeyError:
                pass
            # This needs to be `reraise()`, and not just `raise`. Otherwise,
            # the inner exception from the previous line is re-raised, which
            # isn't desired.
            reraise(*exc_info)

    def __iter__(cls):
        return chain(super(ExtendableEnumMeta, cls).__iter__(), chain.from_iterable(map(iter, cls.__subclasses__())))

    def __len__(cls):
        return super(ExtendableEnumMeta, cls).__len__() + sum(map(len, cls.__subclasses__()))

    def __reversed__(cls):
        return reversed(list(cls))
